"""
In this tool we use the Model View Controller (MVC) design pattern.
The MCV can be considered an approach to distinguish between the data model, processing control and the user interface.
It neatly separates the graphical interface displayed to the user from the code that manages the user actions.

In this file we define the 'Boxmanager_view_ABC' which is the generic representation of the view

Here we define the structure of the view and implement all the function which call the GUI framework and wrapper libs,
at the time of writing (24.1.22) we are using pyqt5 (the try-except statment in the import step will be removed in the future).

In this way if in the future we want to change the wrapper libs we have 'only to change the syntax in this file and in
'helper_GUI.py' without changing the logic of the whole view


"""

try:
    QT = 4
    import PyQt4.QtGui as QtG
    import PyQt4.QtCore as QtCore
    from PyQt4.QtGui import QFontMetrics,QFont
    import matplotlib.backends.backend_qt4agg as plt_qtbackend
except ImportError:
    QT = 5
    import PyQt5.QtWidgets as QtG
    from PyQt5.QtGui import QFontMetrics,QFont
    import PyQt5.QtCore as QtCore
    import matplotlib.backends.backend_qt5agg as plt_qtbackend

from abc import abstractmethod
import cryoloBM.__init__ as ini
from cryoloBM import constants,helper_GUI,boxmanager_toolbar
from typing import Tuple,List
from numpy import ndarray
import matplotlib.pyplot as plt

class Boxmanager_view_ABC(QtG.QMainWindow):
    def __init__(self, font:QFont)->None:
        self.font = font
        super(Boxmanager_view_ABC, self).__init__(None)
        self.current_image_path = None
        self.image_folder_path = None

        # when we change the combobox layout the 'currentIndexChanged' is triggered. We have to avoid to run the function
        self.is_updating_cb = False

        # it is used for disabled pop up messages when update params and GUI
        self.is_updating_params = False

        # figure and axes var
        self.fig, self.ax = None,None

        # self.fig.canvas
        self.background_current = None
        self.background_current_onmove = None       # for managing the real time filament picking

        # plot var
        self.plot = None

        # var to show the image (it ll be self.ax.imshow)
        self.im = None

        # flag for identifying if a slider is pressed
        self.is_slider_pressed = False

        self.show_confidence_histogram_action = QtG.QAction("Confidence histogram", self)
        self.show_size_distribution_action = QtG.QAction("Size distribution", self)

        self.mainMenu = self.menuBar()
        self.fileMenu = self.mainMenu.addMenu("&File")
        self.plotMenu = self.mainMenu.addMenu("&Plot")

        self.layout = None
        self.tree = QtG.QTreeWidget(self)
        self.current_tree_item = None

        self.zoom_update = False

        self.preview_win = None     # The preview window
        self.preview_is_on = False # True when the preview win (i.e.: tracing results) is displayed

        # Init params and related Qt object used in the tabs
        # VISUALIZATION TAB
        self.boxsize_line = QtG.QLineEdit(str(constants.Default_settings_value.DEFAULT_BOX_SIZE.value))
        self.button_set_box_size = QtG.QPushButton("Set")
        self.boxsize_label = QtG.QLabel("Box size: ")

        self.visualization_label = QtG.QLabel("Visualization: ")

        self.visualization_combobox = QtG.QComboBox()

        # the box distance is visible only when we are picking filaments. i.e.. Picking combobox set on FILAMENT
        self.box_distance_filament_line = QtG.QLineEdit(str(constants.Default_settings_value.DEFAULT_BOX_DISTANCE_FILAMENT_SIZE.value))
        self.button_set_box_distance_filament = QtG.QPushButton("Set")
        self.box_distance_filament_label = QtG.QLabel("Box distance: ")
        self.button_set_box_distance_filament.setVisible(False)
        self.box_distance_filament_label.setVisible(False)
        self.box_distance_filament_line.setVisible(False)

        self.use_estimated_size_label = QtG.QLabel("Use estimated size:")
        self.use_estimated_size_checkbox = QtG.QCheckBox()

        self.picking_label = QtG.QLabel("Picking:")
        self.picking_filament_combox = QtG.QComboBox()

        # THRESHOLDING TAB
        self.lower_size_thresh_label = QtG.QLabel("Minimum size: ")
        self.lower_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.lower_size_thresh_slide = QtG.QSlider(QtCore.Qt.Horizontal)
        self.lower_size_thresh_line = QtG.QLineEdit(
            str(constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value))

        self.upper_size_thresh_label = QtG.QLabel("Maximum size: ")
        self.upper_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value
        self.upper_size_thresh_slide = QtG.QSlider(QtCore.Qt.Horizontal)
        self.upper_size_thresh_line = QtG.QLineEdit(
            str(constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value))

        self.current_conf_thresh = constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value
        self.conf_thresh_label = QtG.QLabel("Confidence threshold: ")
        self.conf_thresh_slide = QtG.QSlider(QtCore.Qt.Horizontal)
        self.conf_thresh_line = QtG.QLineEdit(str(constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value))

        self.current_num_boxes_thresh = constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value
        self.num_boxes_thres_label = QtG.QLabel("Number boxes threshold: ")
        self.num_boxes_thresh_slide = QtG.QSlider(QtCore.Qt.Horizontal)
        self.num_boxes_thresh_line = QtG.QLineEdit(
            str(constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value))

        # FILTERING TAB
        self.filter_freq = constants.Default_settings_value.DEFAULT_FILTER_FREQ.value
        self.filter_line = QtG.QLineEdit(str(self.filter_freq))
        self.low_pass_filter_label = QtG.QLabel("Low pass filter cut-off: ")
        self.button_apply_filter = QtG.QPushButton("Apply")
        self.janny_label = QtG.QLabel("Janny denoising: ")
        self.button_janny = QtG.QPushButton("Run")

        # TRACING TAB
        self.preview_label = QtG.QLabel("Preview")
        self.preview_checkbox = QtG.QCheckBox()
        self.button_trace = QtG.QPushButton("Trace")

        self.search_range = constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value
        self.search_range_line = QtG.QLineEdit(str(self.search_range))
        self.search_range_label = QtG.QLabel("Search range: ")
        self.search_range_slider = QtG.QSlider(QtCore.Qt.Horizontal)

        self.memory = constants.Default_settings_value.DEFAULT_MEMORY.value
        self.memory_line = QtG.QLineEdit(str(self.memory))
        self.memory_label = QtG.QLabel("Memory: ")
        self.memory_slider = QtG.QSlider(QtCore.Qt.Horizontal)

        self.min_length = constants.Default_settings_value.DEFAULT_MIN_LENGTH.value
        self.min_length_line = QtG.QLineEdit(str(self.min_length))
        self.min_length_label = QtG.QLabel("Minimum length: ")
        self.min_length_slider = QtG.QSlider(QtCore.Qt.Horizontal)

        # param for tracking the filament
        self.win_size = None  # it will set for default to 'self.box_size'
        self.win_size_line = QtG.QLineEdit()
        self.win_size_label = QtG.QLabel("Window size: ")
        self.win_size_slider = QtG.QSlider(QtCore.Qt.Horizontal)

        self.min_edge_weight = constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value
        self.min_edge_weight_line = QtG.QLineEdit(str(self.min_edge_weight))
        self.min_edge_weight_label = QtG.QLabel("Minimum edge weight: ")
        self.min_edge_weight_slider = QtG.QSlider(QtCore.Qt.Horizontal)

        # add tabs
        self.tabs = QtG.QTabWidget()
        self.tab_visualization = QtG.QWidget()
        self.tab_thresholding = QtG.QWidget()
        self.tab_tracing = QtG.QWidget()
        self.tab_tracing.setToolTip("THE TRACING IS AVAILABLE ONLY FOR TOMO ")
        self.tab_filtering = QtG.QWidget()

    def setup_(self)->None:
        # SETUP QT
        self.setWindowTitle("BoxManager " + ini.__version__)
        central_widget = QtG.QWidget(self)

        self.setCentralWidget(central_widget)

        # Center on screen
        resolution = QtG.QDesktopWidget().screenGeometry()
        self.move(
            (resolution.width() / 2) - (self.frameSize().width() / 2),
            (resolution.height() / 2) - (self.frameSize().height() / 2),
            )

        # Setup Menu
        close_action = QtG.QAction("Close", self)
        close_action.setShortcut("Ctrl+Q")
        close_action.setStatusTip("Leave the app")
        close_action.triggered.connect(self.close_boxmanager)

        open_image_folder = QtG.QAction("Folder", self)
        open_image_folder.triggered.connect(self.open_SPA_folder)     # it was open_image_folder

        open_image = QtG.QAction("File", self)
        open_image.triggered.connect(self.open_SPA_image)

        open_image_3D_folder = QtG.QAction("Folder", self)
        open_image_3D_folder.triggered.connect(self.open_tomo_folder) #it was open_image3D_folder

        open_image3D_tomo = QtG.QAction("File", self)
        open_image3D_tomo.triggered.connect(self.open_tomo_image)       # it was open_image3D_tomo

        import_box_folder = QtG.QAction("Import box files", self)
        import_box_folder.triggered.connect(self.import_box_files)        # ii was load_box_files

        save_data = QtG.QAction("Save", self)
        save_data.triggered.connect(self.save_on_files)  #it was write_all_type

        resetMenu = QtG.QAction("Reset", self)
        resetMenu.triggered.connect(self.reset_config)

        self.show_confidence_histogram_action.triggered.connect( self.show_confidence_histogram)
        self.show_confidence_histogram_action.setEnabled(False)

        self.show_size_distribution_action.triggered.connect(self.show_size_distribution)
        self.show_size_distribution_action.setEnabled(False)

        openMenu = self.fileMenu.addMenu("&Open")
        openMenuSPA = openMenu.addMenu("&SPA")
        openMenuSPA.addAction(open_image_folder)
        openMenuSPA.addAction(open_image)
        openMenuTomo = openMenu.addMenu("&Tomogram")
        openMenuTomo.addAction(open_image_3D_folder)
        openMenuTomo.addAction(open_image3D_tomo)

        self.fileMenu.addAction(import_box_folder)
        self.fileMenu.addAction(save_data)
        self.fileMenu.addAction(resetMenu)

        self.fileMenu.addAction(close_action)

        self.plotMenu.addAction(self.show_confidence_histogram_action)
        self.plotMenu.addAction(self.show_size_distribution_action)

        # Setup tree
        self.setMenuBar(self.mainMenu)

        self.layout = QtG.QGridLayout(central_widget)

        self.tree.setHeaderHidden(True)
        self.layout.addWidget(self.tree, 0, 0, 1, 3)
        self.tree.currentItemChanged.connect(self.event_image_changed)
        self.tree.itemChanged.connect(helper_GUI.event_checkbox_changed)

        self.tabs.addTab(self.tab_visualization, "Visualization")
        self.tabs.addTab(self.tab_thresholding, "Thresholding")
        self.tabs.addTab(self.tab_filtering, "Filtering")

        self.create_tab_visualization()
        self.create_tab_filtering()
        self.create_tab_thresholding()
        #self.create_tab_tracing()      #todo uncomment it for seeing the tracing tab

        #add tabs to widget
        line_counter = 1
        # picking_filament
        self.picking_label.setEnabled(True)
        self.layout.addWidget(self.picking_label, line_counter, 0)
        line_counter = line_counter + 1
        self.picking_filament_combox.addItems([constants.Picking_cb.PARTICLE.value,constants.Picking_cb.FILAMENT.value])
        self.picking_filament_combox.setEnabled(True)
        self.picking_filament_combox.currentIndexChanged.connect(self.picking_filament_changed)
        self.layout.addWidget(self.picking_filament_combox, line_counter, 0)
        line_counter = line_counter + 1

        self.layout.addWidget(self.tabs, line_counter, 0)

        # Show image selection
        self.show()

    @abstractmethod
    def create_tab_filtering(self)->None:
        line_counter = 1
        layout = QtG.QGridLayout()

        #todo: uncomment these lines when the janny option will be implemented
        """
        It is still not implemented and we make it invisibl 
        #run janny
        self.janny_label.setEnabled(True)
        layout.addWidget(self.janny_label, line_counter, 0)
        self.button_janny.clicked.connect(self.janny)
        layout.addWidget(self.button_janny, line_counter, 1)
        line_counter += 1
        """

        # Low pass filter setup
        self.low_pass_filter_label.setEnabled(True)
        layout.addWidget(self.low_pass_filter_label, line_counter, 0)
        layout.addWidget(self.filter_line, line_counter, 1)
        self.filter_line.returnPressed.connect(self.apply_filter)
        self.button_apply_filter.clicked.connect(self.apply_filter)
        self.button_apply_filter.setEnabled(False)
        layout.addWidget(self.button_apply_filter, line_counter, 2)

        self.tab_filtering.setLayout(layout)

    @abstractmethod
    def create_tab_visualization(self)->None:
        """
        It creates the 'visualization' tab used for changing the visualization of the particles/filaments
        this tab is created at the run.
        """
        line_counter = 1
        layout = QtG.QGridLayout()

        # Box size setup
        layout.addWidget(self.boxsize_label, line_counter, 0)
        self.boxsize_line.returnPressed.connect(self.box_size_changed)
        layout.addWidget(self.boxsize_line, line_counter, 1)
        self.button_set_box_size.clicked.connect(self.box_size_changed)
        layout.addWidget(self.button_set_box_size, line_counter, 2)
        line_counter = line_counter + 1

        # Use circle instead of rectangle
        self.visualization_label.setEnabled(True)
        layout.addWidget(self.visualization_label, line_counter, 0)
        self.visualization_combobox.addItems([constants.Visualization_cb.RECT.value, constants.Visualization_cb.CIRCLE.value])
        self.visualization_combobox.currentIndexChanged.connect(self.visualization_changed)
        self.visualization_combobox.setEnabled(True)
        layout.addWidget(self.visualization_combobox, line_counter, 1)
        line_counter = line_counter + 1

        # box picking distance
        layout.addWidget(self.box_distance_filament_label, line_counter, 0)
        self.box_distance_filament_line.returnPressed.connect(self.box_distance_filament_changed)
        layout.addWidget(self.box_distance_filament_line, line_counter, 1)
        self.button_set_box_distance_filament.clicked.connect(self.box_distance_filament_changed)
        layout.addWidget(self.button_set_box_distance_filament, line_counter, 2)
        #self.is_box_distance_filament_visible(is_visible=False)

        line_counter = line_counter + 1

        # Show estimated size
        self.use_estimated_size_label.setEnabled(False)
        layout.addWidget(self.use_estimated_size_label, line_counter, 0)
        layout.addWidget(self.use_estimated_size_checkbox, line_counter, 1)
        self.use_estimated_size_checkbox.stateChanged.connect(self.use_estimated_size_changed)
        self.use_estimated_size_checkbox.setEnabled(False)
        self.tab_visualization.setLayout(layout)

    @abstractmethod
    def create_tab_thresholding(self)->None:
        """
        It creates the 'thresholding' tab used for filtering the particles/filaments got via Cryolo
        this tab is created at the run. Its variables will be unblur after after loading cbox traced/untraced file
        """
        line_counter = 1
        layout = QtG.QGridLayout()

        # Lower size
        layout.addWidget(self.lower_size_thresh_label, line_counter, 0)
        self.lower_size_thresh_slide.setMinimum(constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value)
        self.lower_size_thresh_slide.setMaximum(500)
        self.lower_size_thresh_slide.setValue(constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value)
        self.lower_size_thresh_slide.valueChanged.connect(self.lower_size_thresh_changed)
        self.lower_size_thresh_slide.sliderPressed.connect(self.slider_pressed)
        self.lower_size_thresh_slide.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.LOWER_SIZE.value))
        self.lower_size_thresh_slide.setTickPosition(QtG.QSlider.TicksBelow)
        self.lower_size_thresh_slide.setTickInterval(1)
        layout.addWidget(self.lower_size_thresh_slide, line_counter, 1)

        self.lower_size_thresh_line.textChanged.connect(self.lower_size_label_changed)
        layout.addWidget(self.lower_size_thresh_line, line_counter, 2)
        self.lower_size_thresh_line.returnPressed.connect(self.lower_size_label_changed)

        line_counter = line_counter + 1

        # Upper size threshold
        layout.addWidget(self.upper_size_thresh_label, line_counter, 0)
        self.upper_size_thresh_slide.setMinimum(0)
        self.upper_size_thresh_slide.setMaximum(constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value)
        self.upper_size_thresh_slide.setValue(constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value)
        self.upper_size_thresh_slide.valueChanged.connect(self.upper_size_thresh_changed)
        self.upper_size_thresh_slide.sliderPressed.connect(self.slider_pressed)
        self.upper_size_thresh_slide.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.UPPER_SIZE.value))
        self.upper_size_thresh_slide.setTickPosition(QtG.QSlider.TicksBelow)
        self.upper_size_thresh_slide.setTickInterval(1)
        layout.addWidget(self.upper_size_thresh_slide, line_counter, 1)
        self.upper_size_thresh_line.textChanged.connect(self.upper_size_label_changed)
        layout.addWidget(self.upper_size_thresh_line, line_counter, 2)
        self.upper_size_thresh_line.returnPressed.connect(self.upper_size_label_changed)

        line_counter = line_counter + 1

        # Confidence threshold setup
        layout.addWidget(self.conf_thresh_label, line_counter, 0)
        self.conf_thresh_slide.setMinimum(0)
        self.conf_thresh_slide.setMaximum(100)
        self.conf_thresh_slide.setValue(30)
        self.conf_thresh_slide.valueChanged.connect(self.conf_thresh_changed)
        self.conf_thresh_slide.sliderPressed.connect(self.slider_pressed)
        self.conf_thresh_slide.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.CONF.value))
        self.conf_thresh_slide.setTickPosition(QtG.QSlider.TicksBelow)
        self.conf_thresh_slide.setTickInterval(1)
        layout.addWidget(self.conf_thresh_slide, line_counter, 1)
        self.conf_thresh_line.textChanged.connect(self.conf_thresh_label_changed)
        layout.addWidget(self.conf_thresh_line, line_counter, 2)
        self.conf_thresh_line.returnPressed.connect(self.conf_thresh_label_changed)

        line_counter = line_counter + 1

        # number of boxes threshold setup
        layout.addWidget(self.num_boxes_thres_label, line_counter, 0)
        self.num_boxes_thresh_slide.setMinimum(constants.Default_settings_value.DEFAULT_MIN_NUM_BOXES_THRESH.value)
        self.num_boxes_thresh_slide.setMaximum(constants.Default_settings_value.DEFAULT_MAX_NUM_BOXES_THRESH.value)
        self.num_boxes_thresh_slide.setValue(constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value)
        #self.num_boxes_thresh_slide.valueChanged.connect(self.num_boxes_thresh_changed)
        #self.num_boxes_thresh_slide.sliderPressed.connect(self.slider_pressed)
        #self.num_boxes_thresh_slide.sliderReleased.connect(self.changed_slider_release_num_boxes_thresh)
        self.num_boxes_thresh_slide.setTickPosition(QtG.QSlider.TicksBelow)
        self.num_boxes_thresh_slide.setTickInterval(1)
        layout.addWidget(self.num_boxes_thresh_slide, line_counter, 1)
        #self.num_boxes_thresh_line.textChanged.connect(self.num_boxes_thresh_label_changed)
        layout.addWidget(self.num_boxes_thresh_line, line_counter, 2)
        # self.num_boxes_thresh_line.returnPressed.connect(self.num_boxes_thresh_label_changed)

        #helper_GUI.thresholding_tab_blurring(controller, is_enable=False, kind_of_cbox=False, has_filament=False)
        self.tab_thresholding.setLayout(layout)

    @abstractmethod
    def create_tab_tracing(self)->None:
        """
        this tab is not created at the run of the GUI but it will appear automatically after loading cbox untraced file
        At the time of writing (24.1.22) this option is enable only for the particle
        """
        # when it ll be usable we ll let blurred some options in not filament case
        self.tabs.addTab(self.tab_tracing, "Tracing")
        line_counter = 1
        layout = QtG.QGridLayout()

        #run tracing_searchRange
        layout.addWidget(self.search_range_label, line_counter, 0)
        self.search_range_slider.setMinimum(0)
        self.search_range_slider.setMaximum(1000)   # it ll be change in real time when we change tomo
        self.search_range_slider.setValue(constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value)
        self.search_range_slider.valueChanged.connect(self.searchRange_changed)
        self.search_range_slider.sliderPressed.connect(self.slider_pressed)
        self.search_range_slider.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.SEARCH_RANGE.value))
        self.search_range_slider.setTickPosition(QtG.QSlider.TicksBelow)
        self.search_range_slider.setTickInterval(1)
        layout.addWidget(self.search_range_slider, line_counter, 1)
        self.search_range_line.textChanged.connect(self.searchRange_label_changed)
        self.search_range_line.returnPressed.connect(self.searchRange_label_changed)
        layout.addWidget(self.search_range_line, line_counter, 2)
        line_counter += 1

        #run tracing_memory
        layout.addWidget(self.memory_label, line_counter, 0)
        self.memory_slider.setMinimum(0)
        self.memory_slider.setMaximum(1000)     # it ll be change in real time when we change tomo
        self.memory_slider.setValue(constants.Default_settings_value.DEFAULT_MEMORY.value)
        self.memory_slider.valueChanged.connect(self.memory_changed)
        self.memory_slider.sliderPressed.connect(self.slider_pressed)
        self.memory_slider.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.MEMORY.value))
        self.memory_slider.setTickPosition(QtG.QSlider.TicksBelow)
        self.memory_slider.setTickInterval(1)
        layout.addWidget(self.memory_slider, line_counter, 1)
        self.memory_line.textChanged.connect(self.memory_label_changed)
        self.memory_line.textChanged.connect(self.memory_label_changed)
        layout.addWidget(self.memory_line, line_counter, 2)
        line_counter += 1

        #run tracing_min_length
        layout.addWidget(self.min_length_label, line_counter, 0)
        self.min_length_slider.setMinimum(0)
        self.min_length_slider.setMaximum(1000)    # it ll be change in real time when we change tomo
        self.min_length_slider.setValue(constants.Default_settings_value.DEFAULT_MIN_LENGTH.value)
        self.min_length_slider.valueChanged.connect(self.min_length_changed)
        self.min_length_slider.sliderPressed.connect(self.slider_pressed)
        self.min_length_slider.sliderReleased.connect(lambda: self.changed_slider_release( type_call= constants.Slider_release_fcall.MIN_LENGTH.value))
        self.min_length_slider.setTickPosition(QtG.QSlider.TicksBelow)
        self.min_length_slider.setTickInterval(1)
        layout.addWidget(self.min_length_slider, line_counter, 1)
        self.min_length_line.textChanged.connect(self.min_length_label_changed)
        self.min_length_line.textChanged.connect(self.min_length_label_changed)
        layout.addWidget(self.min_length_line, line_counter, 2)
        line_counter += 1

        # usable only in filament case
        layout.addWidget(self.min_edge_weight_label, line_counter, 0)
        self.min_edge_weight_slider.setMinimum(0)
        self.min_edge_weight_slider.setMaximum(100)
        self.min_edge_weight_slider.setValue(constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value * 100)
        self.min_edge_weight_slider.valueChanged.connect(self.min_edge_weight_changed)
        self.min_edge_weight_slider.sliderPressed.connect(self.slider_pressed)
        self.min_edge_weight_slider.sliderReleased.connect(lambda: self.changed_slider_release(type_call= constants.Slider_release_fcall.MIN_EDGE_W.value))
        self.min_edge_weight_slider.setTickPosition(QtG.QSlider.TicksBelow)
        self.min_edge_weight_slider.setTickInterval(1)
        layout.addWidget(self.min_edge_weight_slider, line_counter, 1)
        self.min_edge_weight_line.textChanged.connect(self.min_edge_weight_label_changed)
        self.min_edge_weight_line.textChanged.connect(self.min_edge_weight_label_changed)
        layout.addWidget(self.min_edge_weight_line, line_counter, 2)
        line_counter += 1

        self.win_size = constants.Default_settings_value.DEFAULT_BOX_SIZE.value
        layout.addWidget(self.win_size_label, line_counter, 0)
        self.win_size_line.setText(str(self.win_size))
        self.win_size_slider.setMinimum(0)
        self.win_size_slider.setValue(self.win_size)
        self.min_edge_weight_slider.valueChanged.connect(self.win_size_changed)
        self.min_edge_weight_slider.sliderPressed.connect(self.slider_pressed)
        self.win_size_slider.sliderReleased.connect(lambda: self.changed_slider_release(type_call= constants.Slider_release_fcall.WIN_SIZE.value))
        self.win_size_slider.setTickPosition(QtG.QSlider.TicksBelow)
        self.win_size_slider.setTickInterval(1)
        layout.addWidget(self.win_size_slider, line_counter, 1)
        self.win_size_line.textChanged.connect(self.win_size_label_changed)
        self.win_size_line.textChanged.connect(self.win_size_label_changed)
        layout.addWidget(self.win_size_line, line_counter, 2)
        line_counter += 1

        # run the 3D preview
        layout.addWidget(self.preview_label, line_counter, 0)
        self.preview_checkbox.stateChanged.connect(self.preview)
        layout.addWidget(self.preview_checkbox, line_counter, 1)

        # run the 3D trace
        self.button_trace.clicked.connect(self.trace)
        layout.addWidget(self.button_trace, line_counter, 2)
        self.tracing_tab_blurring(is_enable=False, has_filament=False)
        self.tab_tracing.setLayout(layout)

    @abstractmethod
    def reset_config(self) -> None:
        """
        reset function called after clicking Reset->File
        It is defined in the controller, it calls self.reset_config_ and controller.model.reset_config
        """
        pass

    def reset_config_(self)->None:
        """
        Reset config function called by the controller
        """
        # no reset 'current_image_path', otherwise after resetting we cannot change image
        #self.current_image_path = None
        # reset tab Visualization
        self.boxsize_line.setText(str(constants.Default_settings_value.DEFAULT_BOX_SIZE.value))
        self.box_distance_filament_line.setText(str(constants.Default_settings_value.DEFAULT_BOX_DISTANCE_FILAMENT_SIZE.value))
        self.use_estimated_size_checkbox.setEnabled(False)
        self.use_estimated_size_label.setEnabled(False)

        self.set_picking_combobox(is_enabled=True)
        self.set_visibility_box_distance(is_visible=False)
        self.set_visualization_combobox(type_visualization=constants.Display_visualization_cb.PARTICLE.value)

        # it is disabled after importing TRACED filament
        self.button_set_box_size.setEnabled(True)
        self.boxsize_line.setEnabled(True)
        self.boxsize_label.setEnabled(True)

        # reset tab Tracing
        self.preview_is_on = False
        self.preview_win = None

        self.is_updating_cb = False
        self.is_slider_pressed = False

        self.is_updating_params = True
        self.use_estimated_size_checkbox.setCheckState(QtCore.Qt.Unchecked)
        self.preview_checkbox.setCheckState(QtCore.Qt.Unchecked)
        self.is_updating_params = False

        self.zoom_update = False

        # i restore the clean canvas in case of file->reset
        self.background_current_onmove = self.fig.canvas.copy_from_bbox(self.ax.bbox) if self.fig else None
        self.background_current = self.fig.canvas.copy_from_bbox(self.ax.bbox) if self.fig else None


    @abstractmethod
    def set_picking_combobox(self,  is_enabled: bool) -> None:
        """
        It fills the picking combobox of the main window
        :param is_enabled: False it is blurred
        """
        helper_GUI.set_picking_combobox(self,is_enabled=is_enabled)

    @abstractmethod
    def set_visualization_combobox(self, type_visualization: int) -> None:
        """
        It fills the 'visualization' combobox of the 'visualization' tab in function of the picking combobox value
        :param type_visualization: see constant.Display_visualization_cb.value
        """
        self.is_updating_cb = True
        helper_GUI.set_visualization_combobox(self, type_visualization=type_visualization)
        self.is_updating_cb = False

    @abstractmethod
    def set_visibility_box_distance(self, is_visible: bool) -> None:
        """
        It makes the 'box distance' buttons and line of the 'visualization' tab visible or not
        :param is_visible: True makes visible these widgets
        """
        helper_GUI.set_visibility_box_distance(self,is_visible=is_visible)

    @abstractmethod
    def reset_tree(self, root:QtG, title:str)->None:
        self.current_tree_item = None
        self.tree.clear()
        self.tree.setColumnCount(2)
        self.tree.setHeaderHidden(False)
        self.tree.setHeaderLabels(["Filename", "Number of boxes"])
        if self.plot is not None:
            self.plot.close()
        self.tree.addTopLevelItem(root)
        fm = QFontMetrics(self.font)
        w = fm.width(title)
        self.tree.setMinimumWidth(w + 150)
        self.tree.setColumnWidth(0, 300)

    @abstractmethod
    def slider_pressed(self)->None:
        """
        Set the vars which manage the pressing of the sliders
        """
        self.is_slider_pressed = True
        self.is_updating_params = True

    @abstractmethod
    def changed_slider_release(self,  type_call:constants.Slider_release_fcall)->None:
        """
        Is triggered by releasing a slider
        :param type_call: identify which slider was released
        """
        self.is_slider_pressed = False
        self.is_updating_params = False
        if type_call == constants.Slider_release_fcall.CONF.value:
            self.conf_thresh_changed()
        elif type_call == constants.Slider_release_fcall.LOWER_SIZE.value:
            self.lower_size_thresh_changed()
        elif type_call == constants.Slider_release_fcall.UPPER_SIZE.value:
            self.upper_size_thresh_changed()
        elif type_call == constants.Slider_release_fcall.MEMORY.value:
            self.memory_changed()
        elif type_call == constants.Slider_release_fcall.MIN_EDGE_W.value:
            self.min_edge_weight_changed()
        elif type_call == constants.Slider_release_fcall.MIN_LENGTH.value:
            self.min_length_changed()
        elif type_call == constants.Slider_release_fcall.SEARCH_RANGE.value:
            self.searchRange_changed()
        elif type_call == constants.Slider_release_fcall.WIN_SIZE.value:
            self.win_size_changed()


        self.is_updating_params = False
        # we do not want live tracing for filament case
        # real time visualization and thresholding will be managed directly in their function
        #helper_tab_tracing.check_and_run_preview(controller=controller,only_reload=True)

    def set_first_time_img(self,im:ndarray)->None:
        """
        Set the variable to show an image
        :param im: np array
        :return: none
        """
        # Create figure and axes
        self.fig, self.ax = plt.subplots(1)

        self.ax.xaxis.set_visible(False)
        self.ax.yaxis.set_visible(False)

        self.fig.tight_layout()

        # Display the image
        self.im = self.ax.imshow(im, origin="lower", cmap="gray", interpolation="Hanning", vmin=-5, vmax=5)

        self.plot = helper_GUI.QtG.QDialog(self)
        self.plot.canvas = helper_GUI.plt_qtbackend.FigureCanvasQTAgg(self.fig)
        self.plot.canvas.mpl_connect("button_press_event", self.onclick)
        self.plot.canvas.mpl_connect("key_press_event", self.myKeyPressEvent)
        self.plot.canvas.setFocusPolicy(helper_GUI.QtCore.Qt.ClickFocus)
        self.plot.canvas.setFocus()
        self.plot.canvas.mpl_connect("button_release_event", self.onrelease)
        self.plot.canvas.mpl_connect("motion_notify_event", self.onmove)
        self.plot.canvas.mpl_connect("resize_event", self.onresize)
        self.plot.canvas.mpl_connect("draw_event", self.ondraw)
        self.plot.toolbar = self.create_plot_toolbar()

        layout = helper_GUI.QtG.QVBoxLayout()
        layout.addWidget(self.plot.toolbar)
        layout.addWidget(self.plot.canvas)

        self.plot.setLayout(layout)
        self.plot.canvas.draw()
        self.plot.show()
        self.background_current = self.fig.canvas.copy_from_bbox(self.ax.bbox)

    @abstractmethod
    def create_plot_toolbar(self):
        pass

    @abstractmethod
    def picking_filament_changed(self):
        pass

    @abstractmethod
    def searchRange_label_changed(self):
        pass

    @abstractmethod
    def searchRange_changed(self):
        pass

    @abstractmethod
    def event_image_changed(self):
        pass

    @abstractmethod
    def show_size_distribution(self):
        pass

    @abstractmethod
    def show_confidence_histogram(self):
        pass

    @abstractmethod
    def close_boxmanager(self):
        pass

    @abstractmethod
    def open_SPA_folder(self):
        pass

    @abstractmethod
    def open_SPA_image(self):
        pass

    @abstractmethod
    def open_tomo_folder(self):
        pass

    @abstractmethod
    def open_tomo_image(self):
        pass

    @abstractmethod
    def import_box_files(self):
        pass

    @abstractmethod
    def save_on_files(self):
        pass

    @abstractmethod
    def apply_filter(self):
        pass

    @abstractmethod
    def box_size_changed(self):
        pass

    @abstractmethod
    def box_distance_filament_changed(self):
        pass

    @abstractmethod
    def visualization_changed(self):
        pass

    @abstractmethod
    def use_estimated_size_changed(self):
        pass

    @abstractmethod
    def lower_size_thresh_changed(self):
        pass

    @abstractmethod
    def lower_size_label_changed(self):
        pass

    @abstractmethod
    def upper_size_thresh_changed(self):
        pass

    @abstractmethod
    def upper_size_label_changed(self):
        pass

    @abstractmethod
    def conf_thresh_changed(self):
        pass

    @abstractmethod
    def conf_thresh_label_changed(self):
        pass

    @abstractmethod
    def memory_changed(self):
        pass

    @abstractmethod
    def memory_label_changed(self):
        pass

    @abstractmethod
    def min_length_changed(self):
        pass

    @abstractmethod
    def min_length_label_changed(self):
        pass

    @abstractmethod
    def min_edge_weight_changed(self):
        pass

    @abstractmethod
    def min_edge_weight_label_changed(self):
        pass

    @abstractmethod
    def win_size_changed(self):
        pass

    @abstractmethod
    def win_size_label_changed(self):
        pass

    @abstractmethod
    def preview(self):
        pass

    @abstractmethod
    def trace(self):
        pass

    @abstractmethod
    def onclick(self, event:helper_GUI.plt_qtbackend):
        pass

    @abstractmethod
    def myKeyPressEvent(self, event:helper_GUI.plt_qtbackend):
        pass

    @abstractmethod
    def onrelease(self, event:helper_GUI.plt_qtbackend):
        pass

    @abstractmethod
    def onmove(self, event:helper_GUI.plt_qtbackend):
        pass

    @abstractmethod
    def onresize(self, event:helper_GUI.plt_qtbackend):
        pass

    @abstractmethod
    def ondraw(self, event:helper_GUI.plt_qtbackend):
        pass

    # wrapper for function defined in helperGUI but used in other function.
    @abstractmethod
    def clicked_cancel(self):
        """
        Return the QT value after clicking on Cancel
        """
        return helper_GUI.QtG.QMessageBox.Cancel

    @abstractmethod
    def clicked_yes(self):
        """
        Return the QT value after clicking on Yes
        """
        return helper_GUI.QtG.QMessageBox.Yes

    @abstractmethod
    def warning_retrain(self):
        """
        Show warning message 'Only the file saved in 'CBOX_UNTRACED' are usable to retrain crYOLO'
        """

        helper_GUI.QtG.QMessageBox.warning(self, "Info",
                                       "Only the file saved in 'CBOX_UNTRACED' are usable to retrain crYOLO")

    @abstractmethod
    def checkbox_state_value(self, checked:bool)->QT:
        """
        Returns the checked state value of the checkbox
        :param checked: True returns
        """
        return helper_GUI.QtCore.Qt.Checked if checked else helper_GUI.QtCore.Qt.Unchecked

    @abstractmethod
    def create_progress_dialog(self, title):
        """
        Return progress dialog object
        :param title: title to show in the progress dialog window
        """
        return helper_GUI.QtG.QProgressDialog(title, "Cancel", 0, 100, self)


    @abstractmethod
    def display_default_params(self):
        """
        Set the default vars as at the start
        """
        helper_GUI.display_default_params(view=self)

    @abstractmethod
    def tracing_tab_blurring(self, is_enable:bool, has_filament:bool):
        """
        Blur the tracing table, it happens when the code is elaborating the tracing step
        """
        helper_GUI.tracing_tab_blurring(view=self, is_enable=is_enable, has_filament=has_filament)

    @abstractmethod
    def blur_estimated_size(self):
        """
        Blur the 'estimated size'
        """
        self.use_estimated_size_checkbox.setEnabled(False)
        self.use_estimated_size_label.setEnabled(False)

    @abstractmethod
    def thresholding_tab_blurring_(self, is_enable:bool, kind_of_cbox:bool, has_filament:bool, type_case:int)->None:
        """
        After loading from file we have to unblur some options in thresholding tabs
        :param is_enable: If True enable SOME options. If False disable ALL the options
        :param kind_of_cbox: True if model.is_cbox or model.is_cbox_untracved is True
        :param has_filament: If True we work with filament
        :param type_case: constants.Type_case. value
        """
        helper_GUI.thresholding_tab_blurring_(self, is_enable=is_enable, kind_of_cbox=kind_of_cbox, has_filament=has_filament, type_case=type_case)

    @abstractmethod
    def update_low_up_thresh(self, max_size:int, min_size:int) -> None:
        """
        After loading from file we have to update upper and lower threshold values
        :param max_size: max size value
        :param min_size: min size value
        """
        helper_GUI.update_low_up_thresh(self,max_size=max_size, min_size=min_size)

    @abstractmethod
    def apply_to_all_the_tomo_question(self, name_param:str)->int:
        """
        Apply the changes to all the tomos
        :param name_param: name of the parameter is changed
        """
        return helper_GUI.apply_to_all_the_tomo_question(self,name_param=name_param)

    @abstractmethod
    def err_message(self, msg:str)->None:
        """
        Show an error message
        :param msg: error message to show
        """
        helper_GUI.err_message(self,msg=msg)

    @abstractmethod
    def qt_esixting_folder(self, msg:str)->str:
        """
        shows the pop up question message and returns the inserted answer
        :param msg: message to show
        """
        return helper_GUI.qt_esixting_folder(self,msg=msg)

    @abstractmethod
    def qtmessage(self, msg:str,has_cancel:bool)->int:
        """
        shows the pop up question message and returns the inserted answer
        :param msg: message to show
        :param has_cancel: If True it has 'QtG.QMessageBox.Cancel' instead of 'QtG.QMessageBox.No'
        """
        return helper_GUI.qtmessage(self,msg=msg, has_cancel=has_cancel)

    @abstractmethod
    def get_selected_folder(self, unsaved_changes:bool)->str:
        """
        Set image_folder_path. It is selected by the user via GUI
        :param unsaved_changes:
        :return The selected folder
        """
        return helper_GUI.get_selected_folder(self,unsaved_changes=unsaved_changes)

    @abstractmethod
    def get_selected_file(self, unsaved_changes:bool)->str:
        """
        Set current_image_path. It is selected by the user via GUI
        :param unsaved_changes:
        :return The selected file
        """
        return helper_GUI.get_selected_file(self,unsaved_changes=unsaved_changes)

    @abstractmethod
    def get_inter_box_distance(self, value:int)->Tuple[int,int]:
        """
        Get inter box distance value before saving the filaments
        :param value: default value displayed after opening the window
        :return new value and result
        """
        return QtG.QInputDialog.getInt(self,"Inter-box distance", "Please set the inter-box distance (in pixel) for each filament:", value=   value)

    @abstractmethod
    def uncheck_all_slides(self,type_case:int)->None:
        """
        Uncheck all the checked checkboxes
        :param type_case: constants.Type_case. value
        """
        helper_GUI.uncheck_all_slides(self,type_case=type_case)

    @abstractmethod
    def set_checkstate_tree_leafs(self,item: QtG.QTreeWidgetItem, entries: List[str], state: QtCore.Qt.CheckState) -> None:
        """
        Check or uncheck the checkboxes in the tree
        :param item: controller.view.tree.invisibleRootItem()
        :param entries: list of entries as strings (e.g.: list of slices)
        :param state: helper_GUI.QtCore.Qt.Checked or helper_GUI.QtCore.Qt.Unchecked
        """
        helper_GUI.set_checkstate_tree_leafs(item=item, entries=entries, state=state)

    @abstractmethod
    def create_tree_widget_item(self,title:str)->QtG.QTreeWidgetItem:
        return helper_GUI.QtG.QTreeWidgetItem([title])

    @abstractmethod
    def check_preview_checkbox(self, check):
        self.preview_checkbox.setCheckState(self.checkbox_state_value(checked=check))

    @abstractmethod
    def check_use_estimated_size_checkbox(self, check):
        self.use_estimated_size_checkbox.setCheckState(self.checkbox_state_value(checked=check))

    @abstractmethod
    def check_item_checkbox(self, item, check:bool)->None:
        """
        The filenames/slices's checkboxes
        """
        item.setCheckState(0,self.checkbox_state_value(checked=check))

    @abstractmethod
    def line_setText(self,line:QtG.QLineEdit,value:str )->None:
        """
        Set the value into the line (e.g: boxsize line)
        """
        line.setText(value)

    @abstractmethod
    def plot_distribution(self, fig: plt.Figure) -> None:
        """
        Shows the Plot statistics
        ;param fig: matplot figure
        """
        plot = helper_GUI.QtG.QDialog(self)
        plot.canvas = helper_GUI.plt_qtbackend.FigureCanvasQTAgg(fig)
        layout = helper_GUI.QtG.QVBoxLayout()
        layout.addWidget(plot.canvas)
        plot.setLayout(layout)
        plot.setWindowTitle("Size distribution")
        plot.canvas.draw()
        plot.show()

    @abstractmethod
    def blur_picking_combobox(self, is_importing:bool, has_filament:bool)->None:
        """
        Blur the picking combobox and set the correct value on it
        :param is_importing: True if we are importing boxes
        :param has_filament: True if we loaded or picking filaments
        """
        if is_importing:
            # check and update picking filament combobox and related vars
            if not has_filament and self.picking_filament_combox.currentText() == constants.Picking_cb.FILAMENT.value or \
                    has_filament and self.picking_filament_combox.currentText() == constants.Picking_cb.PARTICLE.value:
                pick = constants.Picking_cb.FILAMENT if has_filament else constants.Picking_cb.PARTICLE
                self.picking_filament_combox.setCurrentText(pick.value)  # it call 'view.picking_filament_combox.currentIndexChanged.connect(controller.picking_filament_changed)'

        # after loading picking is not possible to switch from picking box and filament. you have to reset
        self.picking_filament_combox.setEnabled(False)

    @abstractmethod
    def set_cbox_visualization(self,  has_filament:bool)->None:
        """
        set circle visualization in case of importing boxes from cbox
        :param has_filament: True if we loaded or picking filaments
        """

        if has_filament and self.visualization_combobox.currentText() != constants.Visualization_cb.CIRCLE_SEGMENTED.value:
            self.visualization_combobox.setCurrentText(constants.Visualization_cb.CIRCLE_SEGMENTED.value)
        elif not has_filament and self.visualization_combobox.currentText() != constants.Visualization_cb.CIRCLE.value:
            self.visualization_combobox.setCurrentText(constants.Visualization_cb.CIRCLE.value)

    @abstractmethod
    def enable_button_apply_filter(self)->None:
        self.button_apply_filter.setEnabled(True)

    @abstractmethod
    def setCurrentItem_viewTree(self,is_tomo_folder:bool, number_tomo:int =0, z:int =0)->None:
        """
        Set the current item of the view's tree
        :param is_tomo_folder: True in case we loaded a folder of tomo
        :parma number_tomo: index of the tomo in the tree
        :param z: number of slice
        """
        # child_1 = folder  child_2 = # tomo  child_3 = # of slice
        if is_tomo_folder:
            self.tree.setCurrentItem(self.tree.invisibleRootItem().child(0).child(number_tomo).child(z))
        else:
            self.tree.setCurrentItem(self.tree.invisibleRootItem().child(number_tomo).child(z))

    @abstractmethod
    def setWindowTitle_viewPlot(self,title:str)->None:
        """
        Set the title of the plot window
        """
        self.plot.setWindowTitle(title)

    @staticmethod
    def run_in_same_thread():
        QtCore.QCoreApplication.instance().processEvents()

    @staticmethod
    def meta_modifier_is_clicked():
        return helper_GUI.QtG.QApplication.keyboardModifiers() == helper_GUI.QtCore.Qt.MetaModifier

    @staticmethod
    def control_modifier_is_clicked():
        return helper_GUI.QtG.QApplication.keyboardModifiers() == helper_GUI.QtCore.Qt.ControlModifier

    @staticmethod
    def fill_root_childs(root: QtG.QTreeWidgetItem, tot_slices: int, is_folder: bool = False,root_child_index: int = None) -> None:
        """
        Creates all the checkboxes of a given tomogram in the GUI
        :param root: The 'QTreeWidgetItem' obj represent the item list (namely the filenames)
        :param tot_slices:  number of slices present in the tomogram
        :param is_folder: True if we loaded a folder of tomo
        :param root_child_index: index of child in root. It is the item, which represent a tomogram in a folder
        """
        helper_GUI.fill_root_childs(root=root, tot_slices=tot_slices, is_folder=is_folder,root_child_index=root_child_index)

    @abstractmethod
    def get_picking_filament_cb_text(self)->str:
        """
        Return the text of the picking filament combobox
        """
        return self.picking_filament_combox.currentText()

    @abstractmethod
    def get_visualization_cb_text(self)->str:
        """
        Return the text of the visualization combobox
        """
        return self.visualization_combobox.currentText()


import unittest
from cryoloBM.helper_filament import picked_filament
import numpy as np
class MyTestCase(unittest.TestCase):

    def test_resize_height_offset_is_normalized(self):
        fil = picked_filament(30)

        fil.begin_fil = [0, 0]
        fil.end_fil = [100, 100]

        fil.set_line_params()
        self.assertAlmostEqual(np.linalg.norm(fil.offset), 1.0)

    def test_resize_height_rect_coords_get_updated_norotate(self):
        fil = picked_filament(30)

        fil.begin_fil = [0, 0]
        fil.end_fil = [0, 100]

        fil.set_line_params()
        sketch = fil.get_rect_sketch()
        fil.height = 50
        sketch_newhieght = fil.get_rect_sketch()

        self.assertAlmostEqual(25, sketch_newhieght.get_xy()[0])

    def test_resize_height_rect_coords_get_updated_rotate90(self):
        fil = picked_filament(30)

        fil.begin_fil = [0, 0]
        fil.end_fil = [100, 0]

        fil.set_line_params()
        fil.height = 50
        sketch_newhieght = fil.get_rect_sketch()

        self.assertAlmostEqual(-25, sketch_newhieght.get_xy()[1])



if __name__ == '__main__':
    unittest.main()



import unittest
from cryoloBM.helper_filament import picked_filament
from cryoloBM.helper_image import is_click_into_box

def create_fil(boxsize,begin,end):
    fil = picked_filament(box_size=boxsize)
    fil.begin_fil = begin
    fil.end_fil = end
    fil.set_line_params()
    return fil.get_rect_sketch()


class vertical_down_up(unittest.TestCase):
    # vertical filament picked from down to up
    boxsize = 200
    fil = create_fil(boxsize=boxsize,begin=[771.5451467268621, 457.17945823927755],end= [790.0372460496613, 3378.9311512415347])

    def test_clicked_out(self):
        self.assertFalse( is_click_into_box(fil=self.fil,x=1206,y=1853,boxsize=self.boxsize,is_tomo=False))

    def test_clicked_in(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=762, y=1492, boxsize=self.boxsize, is_tomo=False))

class vertical_up_down(unittest.TestCase):
    # vertical filament picked from up to down
    boxsize = 200
    fil = create_fil(boxsize=boxsize,begin=[2112.2223476297963, 3351.193002257336],end= [2065.992099322799, 762.2990970654627])

    def test_clicked_out(self):
        self.assertFalse( is_click_into_box(fil=self.fil,x=2620,y=2139,boxsize=self.boxsize,is_tomo=False))

    def test_clicked_in(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=2093, y=2093, boxsize=self.boxsize, is_tomo=False))


class horizontal_left_right(unittest.TestCase):
    # horizontal filament picked left to right
    boxsize = 200
    fil = create_fil(boxsize=boxsize, begin=[429.44130925507886, 3064.5654627539498],
                     end=[3147.779909706546, 3036.827313769751])

    def test_clicked_out_up(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=2056, y=3746, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_down(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=1853, y=2685, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_right(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=3480, y=3064, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_left(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=179, y=3120, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_center(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=1779, y=3073, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_left(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=854, y=3046, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_right(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=2851, y=3009, boxsize=self.boxsize, is_tomo=False))

class horizontal_right_left(unittest.TestCase):
    # horizontal filament picked right to left
    boxsize = 200
    fil = create_fil(boxsize=boxsize, begin=[2953.6128668171555, 725.3148984198644],
                     end=[484.91760722347624, 734.560948081264])

    def test_clicked_out_up(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=1733, y=1095, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_down(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=1760, y=336, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_right(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=3499, y=706, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_left(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=198, y=753, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_center(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=1760, y=734, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_left(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=743, y=716, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_right(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=2546, y=716, boxsize=self.boxsize, is_tomo=False))


class diagonal_left_up_right_down(unittest.TestCase):
    # horizontal filament from left up to right down right
    boxsize = 200
    fil = create_fil(boxsize=boxsize, begin=[3536.1139954853265, 3582.3442437923245],
                     end= [762.2990970654627, 1002.6963882618508])

    def test_clicked_out_up(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=1751, y=2824, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_down(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=2454, y=1927, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_right(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=3582, y=3702, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_left(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=679, y=910, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_center(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=2065, y=2213, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_left(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=2384, y=2421, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_right(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=3027, y=3110, boxsize=self.boxsize, is_tomo=False))



class diagonal_right_up_left_down(unittest.TestCase):
    # horizontal filament from right up to left down right
    boxsize = 200
    fil = create_fil(boxsize=boxsize, begin=[336.9808126410834, 3221.748306997742],
                     end=[3240.2404063205418, 836.267494356659])

    def test_clicked_out_up(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=2121, y=2250, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_down(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=1169, y=1686, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_right(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=3508, y=669, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_out_left(self):
        self.assertFalse(is_click_into_box(fil=self.fil, x=124, y=3378, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_center(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=1705, y=2180, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_left(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=438, y=3083, boxsize=self.boxsize, is_tomo=False))

    def test_clicked_in_right(self):
        self.assertTrue(is_click_into_box(fil=self.fil, x=3046, y=1021, boxsize=self.boxsize, is_tomo=False))


if __name__ == '__main__':
    unittest.main()



"""
In this file i declare all the logic related to the dropdown menu calls in order to improve the readibility of the
boxmanager_controller.py code letting into it only the calls to the function (of course functions like 'reset_config'
that have only fews line of code are, for now, still into the file)
"""

from cryoloBM import constants, helper, helper_GUI, helper_image
from os import path
from mrcfile import mmap as mrcfile_mmap        #it is not in the setup.py because it is already installed by cryolo
from numpy import min as np_min, max as np_max, isnan as np_isnan, sum as np_sum
import matplotlib.pyplot as plt

from cryoloBM import boxmanager_controller,MySketch # for type hinting purpose
from typing import List,Tuple

def close_boxmanager(controller:boxmanager_controller)->None:
    """
    It is called via File->close
    Close the boxmanager script.
    :param controller: Boxmanager_controller obj
    """
    if controller.model.unsaved_changes and controller.view.qtmessage(
                msg="All loaded boxes are discarded. Are you sure?", has_cancel=True) == controller.view.clicked_cancel():
            return
    controller.view.close()

def reset_config(controller:boxmanager_controller)->None:
    """
    It is called via File->Reset
    Reset all the variables of the model and restore the GUI starting layout
    :param controller: Boxmanager_controller obj
    """
    # We draw and remove the sketches in real time in 'event_image_changed' using controller.model.rectangles var
    # we have to remove ONLY the sketches in the current image/slice
    if controller.model.box_dictionary:
        # delete sketches from the figure
        helper_image.delete_all_patches(controller=controller)
        # update the plot for avoiding weird behaviour (matplotlib tutorial suggests to delete a plot and plot a new one)
        controller.view.fig.canvas.draw()

    # the model.box_dictionary.particle_dictionary war ll be clean and as consequence when we ll load another image no
    # sketches ll be drawn
    controller.model.reset_config()
    controller.view.reset_config_()     # it is defined in the abstract class , '_' becuase 'reset_config' is triggered by signal and call this function
    controller.view.uncheck_all_slides(type_case=controller.model.type_case)
    type_case = controller.model.type_case if controller.model.type_case else None
    controller.view.thresholding_tab_blurring(is_enable=False, kind_of_cbox=False, has_filament=False,type_case=type_case)
    controller.view.tracing_tab_blurring(is_enable=False,has_filament=controller.model.has_filament)
    controller.view.display_default_params()

    # model.reset_config cleaned the Box_dictionary.box_dictionary. Hence update_tree_boxes finds 0 boxes in every image
    controller.view.update_tree_boxsizes(update_current=False,all_tomos=False)

    tot_child = 1 if controller.model.type_case != constants.Type_case.TOMO_FOLDER.value else len(controller.model.params)
    for child_id in range(tot_child):
        controller.view.update_global_counter(childid=child_id)

def open_tomo_image(controller:boxmanager_controller,via_cli:bool)->bool:
    """
    Reads a single tomo image and fills the model's var.
    :param controller: Boxmanager_controller obj
    :param via_cli: True if we run from command line
    :return False if something went wrong
    """
    # when the script is run via cli this var is already filled
    if not via_cli:
        reset_config(controller=controller)
        selected_image=controller.view.get_selected_file(unsaved_changes=controller.model.unsaved_changes)
        helper.set_current_image_path(controller=controller, current_image_path=selected_image)

        if not controller.view.current_image_path:
            return False

        if not controller.view.current_image_path.endswith(("mrc", "mrcs", "rec")):
            controller.view.err_message(msg=f"ERROR: The image '{controller.view.current_image_path}' has an invalid format. Must be in .mrc or .mrcs format")
            return False

    controller.model.set_tomo_vars()
    controller.model.current_tomoimage_mmap = helper.read_image(file_path=controller.view.current_image_path, use_mmap=True)

    if len(controller.model.current_tomoimage_mmap.shape) != 3:
        controller.view.err_message(msg=f"ERROR: The image '{controller.view.current_image_path}' is not a 3D image")
        return False

    root = controller.view.create_tree_widget_item( title=controller.view.current_image_path)
    controller.view.check_item_checkbox(item=root, check=False)

    controller.model.type_case = constants.Type_case.NOT_LOADED.value   # avoid the reset tree to run the helper_image.event_image_changed function and crash

    controller.view.reset_tree(root=root, title=controller.view.current_image_path)

    if controller.model.current_tomoimage_mmap.shape[0] > 0:
        purefilename = controller.get_current_filename(with_extension=False)
        controller.model.tot_frames.update({purefilename: controller.model.current_tomoimage_mmap.shape[0]})
        controller.model.smallest_image_dim.update({purefilename: min(controller.model.current_tomoimage_mmap.data.shape[1], controller.model.current_tomoimage_mmap.data.shape[2])})
        controller.view.memory_slider.setMaximum(controller.model.tot_frames[purefilename] - 1)
        controller.view.min_length_slider.setMaximum(controller.model.tot_frames[purefilename] - 1)
        controller.view.search_range_slider.setMaximum(controller.model.smallest_image_dim[purefilename] - 1)
        controller.view.fill_root_childs(root=root, tot_slices=controller.model.current_tomoimage_mmap.shape[0], is_folder=False)

    else:
        controller.view.err_message(msg=f"ERROR: The image '{controller.view.current_image_path}' is recognized as a 3D image but has no slice (i.e.: mrc.data.shape[0]<1")
        return False

    root.setExpanded(True)

    # Show first image
    img_data= controller.model.current_tomoimage_mmap[int(root.child(0).text(0)), :, :]
    controller.view.current_tree_item = root.child(0)

    controller.model.set_rectangles ([])

    controller.model.type_case = constants.Type_case.TOMO.value
    controller.model.init_box_dictionary(onlyfiles=[controller.model.current_image_path],
                                         tot_slices=controller.model.current_tomoimage_mmap.shape[0])
    controller.view.set_first_time_img( im=img_data)
    controller.view.setWindowTitle_viewPlot(title=f"{controller.get_current_filename(with_extension=True)} \tslice: 0" )
    controller.view.setCurrentItem_viewTree(is_tomo_folder=False, number_tomo=0, z=0)

    controller.view.enable_button_apply_filter()
    controller.model.unsaved_changes = False
    return True

def open_tomo_folder(controller:boxmanager_controller, via_cli:bool)->bool:
    """
    Reads the image folder and fills the model's var.
    :param controller: Boxmanager_controller obj
    :param via_cli: True if we run from command line
    :return False if something went wrong
    """
    # when the script is run via cli this var is already filled
    if not via_cli:
        reset_config(controller=controller)
        controller.model.image_folder_path=controller.view.get_selected_folder(unsaved_changes=controller.model.unsaved_changes)
        controller.view.image_folder_path = controller.model.image_folder_path

    controller.model.set_tomo_vars()

    root, onlyfiles, all_items = helper.list_files_in_folder(controller=controller, is_list_tomo=True)

    if not onlyfiles:
        controller.view.err_message( msg=f"ERROR: The folder '{controller.model.image_folder_path}' does not have tomogram")
        return False

    if root and all_items :
        root.setExpanded(True)
        for root_child_index,f in enumerate(onlyfiles):
            purefilename = path.splitext(path.basename(f))[0]
            controller.model.fill_params(list_files=[purefilename])
            with mrcfile_mmap(path.join(controller.model.image_folder_path, f), permissive=True, mode="r") as mrc:
                controller.model.tot_frames.update({purefilename: mrc.data.shape[0]})
                controller.model.smallest_image_dim.update({purefilename: min(mrc.data.shape[1], mrc.data.shape[2])})
                controller.view.memory_slider.setMaximum(controller.model.tot_frames[purefilename] - 1)
                controller.view.min_length_slider.setMaximum(controller.model.tot_frames[purefilename] - 1)
                controller.view.search_range_slider.setMaximum(controller.model.smallest_image_dim[purefilename] - 1)
                if mrc.data.shape[0] > 0:
                    if controller.model.current_tomoimage_mmap is None:
                        helper.set_current_image_path(controller=controller, current_image_path=path.join(controller.model.image_folder_path, f))
                        controller.model.current_tomoimage_mmap = helper.read_image(file_path=controller.model.current_image_path, use_mmap=True)
                        controller.model.type_case = constants.Type_case.TOMO_FOLDER.value
                        controller.model.init_box_dictionary(onlyfiles=onlyfiles, tot_slices=None)
                    controller.model.box_dictionary.init_single_tomo(tomo_name=path.splitext(f)[0],
                                                                     tot_slices=mrc.data.shape[0])
                    controller.view.fill_root_childs(root=root, is_folder=True, root_child_index=root_child_index,
                                                tot_slices=mrc.data.shape[0])
                mrc.close()

        if controller.model.current_tomoimage_mmap  is not None:
            controller.view.set_first_time_img(im=controller.model.current_tomoimage_mmap[0, :, :])
            controller.view.setWindowTitle_viewPlot(title=f"{controller.get_current_filename(with_extension=True)} \tslice: 0")
            controller.view.setCurrentItem_viewTree(is_tomo_folder=True, number_tomo=0, z=0)

            controller.view.enable_button_apply_filter()
            controller.model.unsaved_changes = False
            return True

        controller.view.err_message(msg = f"ERROR: Cannot open '{controller.model.current_image_path}' tomogram")
        return False

def open_SPA_image(controller:boxmanager_controller, via_cli:bool)->bool:
    """
    Reads the image folder and fills the model's var.
    :param controller: Boxmanager_controller obj
    :param via_cli: True if we run from command line
    :return False if something went wrong
    """
    # when the script is run via cli this var is already filled
    if not via_cli:
        reset_config(controller=controller)
        selected_image=controller.view.get_selected_file(unsaved_changes=controller.model.unsaved_changes)
        helper.set_current_image_path(controller=controller, current_image_path=selected_image)

        if not controller.view.current_image_path:
            return False
    controller.model.image_folder_path = path.dirname(controller.view.current_image_path)
    controller.view.image_folder_path = controller.model.image_folder_path
    controller.model.has_3d = False


    fname =path.splitext(path.basename(controller.view.current_image_path))
    fname=fname[0]+fname[1]

    title = path.join(str(controller.model.image_folder_path),controller.model.wildcard) if controller.model.wildcard else str(controller.model.image_folder_path)
    root = controller.view.create_tree_widget_item( title=fname)

    controller.view.check_item_checkbox(item=root,check=False)
    controller.view.reset_tree(root, title)
    controller.model.set_rectangles(rectangles=[])

    all_items = [controller.view.create_tree_widget_item( title=fname)]

    for item_index, item in enumerate(all_items):
        controller.view.run_in_same_thread()
        controller.view.check_item_checkbox(item=item, check=False)
        root.addChild(item)

    root.setExpanded(True)
    # Show first image

    controller.view.current_tree_item = root.child(0)
    im = helper.read_image(file_path=controller.model.current_image_path, use_mmap=False)

    if len(im.shape) != 2:
        controller.view.err_message( msg="Please open an image folder with micrographs")
        return False

    controller.model.set_rectangles(rectangles=[])

    controller.model.type_case = constants.Type_case.SPA.value
    controller.model.init_box_dictionary(onlyfiles=[fname], tot_slices=None)
    controller.view.set_first_time_img( im=im)

    controller.view.setCurrentItem_viewTree(is_tomo_folder=False, number_tomo=0, z=0)
    controller.view.setWindowTitle_viewPlot(title=controller.get_current_filename(with_extension=True))
    controller.view.enable_button_apply_filter()
    controller.model.unsaved_changes = False

    return True

def open_SPA_folder(controller:boxmanager_controller, via_cli:bool)->bool:
    """
    Reads the image folder and fills the model's var.
    :param controller: Boxmanager_controller obj
    :param via_cli: True if we run from command line
    :return False if something went wrong
    """
    # when the script is run via cli this var is already filled
    if not via_cli:
        reset_config(controller=controller)
        controller.model.image_folder_path=controller.view.get_selected_folder(unsaved_changes=controller.model.unsaved_changes)
        controller.view.image_folder_path = controller.model.image_folder_path

    controller.model.has_3d = False

    # do not clean the self.box_dictionary because for the SPA case it ll be overwrite in the 'init_box_dictionary'

    root, onlyfiles, all_items = helper.list_files_in_folder(controller=controller, is_list_tomo=False)

    if not onlyfiles:
        return False

    if root and all_items :
        root.setExpanded(True)
        # Show first image
        helper.set_current_image_path(controller=controller, current_image_path=path.join(controller.model.image_folder_path, str(root.child(0).text(0))))
        controller.view.current_tree_item = root.child(0)
        im = helper.read_image(file_path=controller.model.current_image_path,use_mmap=False)

        if len(im.shape) != 2:
            controller.view.err_message( msg="Please open an image folder with micrographs")
            return False

        controller.model.set_rectangles( rectangles= [])

        controller.model.type_case = constants.Type_case.SPA.value
        controller.model.init_box_dictionary(onlyfiles=onlyfiles, tot_slices=None)
        controller.view.set_first_time_img( im=im)

        controller.view.setCurrentItem_viewTree(is_tomo_folder=False, number_tomo=0, z=0)
        controller.view.setWindowTitle_viewPlot(title=controller.get_current_filename(with_extension=True))
        controller.view.enable_button_apply_filter()
        controller.model.unsaved_changes = False
        return True

    return False

def _get_list_boxes_for_histograms(controller:boxmanager_controller)->Tuple[List[MySketch.MySketch],str]:
    """
    Return the correct list of list of boxes for 'show_confidence_histogram' and 'show_size_distribution'
    """
    current_filename = controller.get_current_filename(with_extension=False)
    add_in_title = " of " + current_filename if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value else ""
    if controller.model.type_case == constants.Type_case.SPA.value:
        list_boxes = [*controller.model.box_dictionary.particle_dictionary.values()]
    else:
        bdict=controller.model.box_dictionary.particle_dictionary[current_filename]
        list_boxes = [*bdict.values()]

    return list_boxes,add_in_title

def show_size_distribution(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    list_boxes, add_in_title = _get_list_boxes_for_histograms(controller=controller)
    estimated_size = [box.get_est_size() for boxes in list_boxes for box in boxes]

    # if there is a Nan value does not plot (sum is 2.5x faster than min)
    if not estimated_size or np_isnan(np_sum(estimated_size)):
        helper_GUI.err_message(view=controller.view, msg="no estimated sizes are available for this dataset")
        return


    fig = plt.figure()

    width = max(10, int((np_max(estimated_size) - np_min(estimated_size)) / 10))
    plt.hist(estimated_size, bins=width)
    plt.title("Particle diameter distribution" + add_in_title)
    plt.xlabel("Partilce diameter [px] (Bin size: " + str(width) + "px )")
    plt.ylabel("Count")

    controller.view.plot_distribution( fig=fig)

def show_confidence_histogram(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    list_boxes, add_in_title = _get_list_boxes_for_histograms(controller)
    confidence= [box.get_confidence() for boxes in list_boxes for box in boxes]

    fig = plt.figure()

    width = max(10, int((np_max(confidence) - np_min(confidence)) / 0.05))
    plt.hist(confidence, bins=width)
    plt.title("Confidence distribution" + add_in_title)
    bin_size_str = "{0:.2f}".format(((np_max(confidence) - np_min(confidence)) / width))
    plt.xlabel("Confidence (Bin size: " + bin_size_str + ")")
    plt.ylabel("Count")

    controller.view.plot_distribution( fig=fig)



#
# This fiel contains command line cryoloBM_tools for box management
#
import sys
import argparse
from typing import List
from cryoloBM.bmtool import BMTool


TYPE_CBOX=0
TYPE_COORDS=1
TYPE_EMAN_BOX=2
TYPE_EMAN_BOX_3D=3
TYPE_EMAN_HELICON=4
TYPE_EMAN_START_END=5
TYPE_RELION_STAR=6

def get_file_type(path):
    from cryolo import CoordsIO
    if CoordsIO.is_eman1_filament_start_end(path):
        return TYPE_EMAN_START_END
    if CoordsIO.is_eman1_helicon(path):
        return TYPE_EMAN_HELICON
    if CoordsIO.is_star_filament_file(path):
        return TYPE_RELION_STAR
    if path.endswith(".coords"):
        return TYPE_COORDS
    if path.endswith(".cbox"):
        return TYPE_CBOX
    return -1


def get_tool_list() -> List[BMTool]:
    tools = []

    from cryoloBM_tools.filamentresampling import FilamentResampleTool
    fil_resample_tool = FilamentResampleTool()
    tools.append(fil_resample_tool)

    from cryoloBM_tools.coords2warp import Coords2WarpTool
    c2w_tool = Coords2WarpTool()
    tools.append(c2w_tool)

    from cryoloBM_tools.priors2star import Priors2StarTool
    p2s_tool = Priors2StarTool()
    tools.append(p2s_tool)

    from cryoloBM_tools.scale import ScaleTool
    scale_tool = ScaleTool()
    tools.append(scale_tool)

    from cryoloBM_tools.rotatecboxcoords import RotateCBOXCoords
    rotate_tool = RotateCBOXCoords()
    tools.append(rotate_tool)

    from cryoloBM_tools.cbox2coords import CBOX2Coords
    cbox2coords_tool = CBOX2Coords()
    tools.append(cbox2coords_tool)

    from cryoloBM_tools.spherical_prior import SphericalPrior
    spherical_prior_tool = SphericalPrior()
    tools.append(spherical_prior_tool)

    from cryoloBM_tools.coords2cbox import Coords2CBOX
    coords2cbox_tool = Coords2CBOX()
    tools.append(coords2cbox_tool)


    return tools

def _main_():

    parser = argparse.ArgumentParser(
        description="Boxmanager Tools",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(help="sub-command help")


    tools = get_tool_list()

    tools = sorted(tools, key=lambda x: x.get_command_name())

    for tool in tools:
        tool.create_parser(subparsers)

    args = parser.parse_args()

    for tool in tools:
        if tool.get_command_name() in sys.argv[1]:
            tool.run(args)

if __name__ == "__main__":
    _main_()


"""
The following functions manage the import data from file option
"""

from cryoloBM import helper_filament, helper_image, constants, helper, MySketch,helper_GUI
from cryolo import CoordsIO,utils

from math import isnan
from os import path,listdir, stat as os_stat
import random
import time

from cryoloBM import boxmanager_controller # for type hinting purpose
from typing import Union, List

COLORS = ["b", "r", "c", "m", "y", "k", "w"]


def box_to_sketch(controller: boxmanager_controller ,box: utils.BoundBox, color)->MySketch.MySketch:
    """
    Convert a bbox obj in a sketch. It is the old 'box_to_rectangle' function
    :param controller: Boxmanager_controller obj
    :param box: the bbox obj
    :param color: the color of the contour of the sketch
    :return : the converted mysketch obj
    """
    is_tomo = controller.model.type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value]
    num_boxes = 1
    # controller.model.boxsize = avg_size # avg_size = (int(box.w) + int(box.h)) // 2 #todo: do i need this????? it is related to helper_GUI.show_loaded_boxes
    est_size = (int(box.w) + int(box.h)) // 2
    meta = None
    if "est_box_size" in box.meta:
        meta = {"est_box_size": box.meta["est_box_size"]}
        est_size = (box.meta["est_box_size"][0] + box.meta["est_box_size"][1]) // 2
    if 'num_boxes' in box.meta and controller.model.is_cbox and isnan(box.meta["num_boxes"]) is False:
        num_boxes = box.meta["num_boxes"]

    confidence = box.c if box.c is not None else 1
    return  MySketch.MySketch(xy =(int(box.x), int(box.y)), width=int(box.w), height=int(box.h), is_tomo=is_tomo,
                              angle=0.0, est_size=est_size, confidence=confidence, only_3D_visualization=False,
                              num_boxes=num_boxes, meta=meta, z=box.z,
                              linewidth=1, edgecolor=color, facecolor="none")

def _remove_old_boxes(controller:boxmanager_controller)->None:
    """
    Delete old boxes and reset boxdictionary's dictionaries
    Called if the user want remove the old boxes:  "Keep old boxes loaded and show the new ones in a different color?"
    :param controller: Boxmanager_controller obj
    """
    # delete these sketches from the figure
    helper_image.delete_all_patches(controller=controller)
    # reset the dictionaries
    controller.model.box_dictionary.reset()
    # uncheck all the slides
    controller.view.uncheck_all_slides(type_case=controller.model.type_case)
    # reset the counters
    controller.view.update_tree_boxsizes( update_current=False, all_tomos=False)
    controller.view.update_global_counter()

    # restore the clean canvas
    if controller.view.background_current:
        controller.view.fig.canvas.restore_region(controller.view.background_current)

def _update_entries_empty_file(f:str , updated_entries:List[str], is_star_startend:bool, is_helicon:bool, is_eman1_startend:bool)->str:
    """
    Update entries for empty file
    if the file is empty i want just to check the related checkbox
    :param f: filename
    :param updated_entries: list of entries to update
    :param is_eman1_startend: True if 'f' is 'eman1_startend' file
    :param is_helicon: True if 'f' is 'is_helicon' file
    :param is_eman1_startend: True if 'f' is 'is_eman1_startend' file
    :return : the entry name
    """
    dict_entry_name = path.splitext(f)[0]
    if is_star_startend:
        dict_entry_name = f[:-5]
    elif is_helicon or is_eman1_startend:
        dict_entry_name = f[:-4]
    updated_entries.append(dict_entry_name)
    return dict_entry_name


def _check_correct_import_data(controller:boxmanager_controller, keep:bool, boxes:List[Union[utils.Filament,utils.BoundBox]])->bool:
    """
    When load data keeping the old data i have to make sure that:
        -) we do not load particle after filament and viceversa
        -) when we are visualizing traced filament we do not load untraced and viceversa (only tomo cases)
    :param controller: Boxmanager_controller obj
    :param keep: True if you will keep the old data
    :param boxes
    :return True if everything ok
    """
    if keep is False:
        return True

    err_msg=""
    if controller.model.has_filament and not isinstance(boxes[0], utils.Filament):
        err_msg="Error: You tried to load particles after loading filaments"
    elif not controller.model.has_filament and isinstance(boxes[0], utils.Filament):
        err_msg = "Error: You tried to load filaments after loading particles"
    elif controller.model.has_filament and controller.model.type_case != constants.Type_case.SPA.value:
        # it is not possible load untraced filament when are visualizing the traced and viceversa
        loading_untraced = helper.is_cbox_untraced(boxes[0].boxes[0])
        if controller.model.is_cbox_untraced and not loading_untraced:
            err_msg ="Error: You tried to load traced filaments after loading untraced filaments"
        if not controller.model.is_cbox_untraced and loading_untraced:
            err_msg = "Error: You tried to load untraced filaments after loading traced filaments"
    else:
        return True

    controller.view.err_message(msg=err_msg)
    return False

def _load_data(controller:boxmanager_controller, pathfile:str, keep:bool, t_start:float)->List[Union[utils.Filament, utils.BoundBox]]:
    """
    :param controller: Boxmanager_controller obj
    :param pathfile: path of the file from which to load data
    :param keep: False if we want to discard the data
    :param t_start: starting time
    """
    boxes = list()
    controller.model.is_cbox = False
    if pathfile.endswith(".box"):
        boxes = CoordsIO.read_eman1_boxfile(path=pathfile, is_SPA=True,box_size_default=200)
    if pathfile.endswith(".star"):
        boxes = CoordsIO.read_star_file(path=pathfile, box_size=controller.model.boxsize)
    if pathfile.endswith(".cbox"):
        boxes = CoordsIO.read_cbox_boxfile(path=pathfile,min_dist_boxes_filament=0)
        print(f"loading {time.time() - t_start}")
        controller.model.is_cbox = True
        if boxes:
            if not _check_correct_import_data(controller=controller,keep=keep, boxes=boxes):
                return list()

            controller.set_has_filament(has_filament= isinstance(boxes[0], utils.Filament))
            b = boxes[0].boxes[0] if controller.model.has_filament else boxes[0]
            sketch = box_to_sketch(controller=controller, box=b, color=None)
            helper_image.box_size_changed(controller=controller, box_size_from_file=sketch.get_width())
            controller.model.est_box_from_cbox = sketch.get_est_size()
            controller.model.box_dictionary.update_boxsize_dictionary(new_boxsize=sketch.get_width(),f=path.splitext(pathfile)[0].split("/")[-1], n=controller.model.index_tomo)

    if (pathfile.endswith(".box") or pathfile.endswith(".star")) and boxes:
        controller.set_has_filament(has_filament = False)
        sketch = box_to_sketch(controller=controller, box=boxes[0], color=None)
        helper_image.box_size_changed(controller=controller, box_size_from_file=sketch.get_width())

    return boxes


def _collect_filament(controller:boxmanager_controller, filaments:List[utils.Filament], dict_entry_name:str, keep:bool)->int:
    """
    Collect the filament, set all the 'filaments'' variables and return the number of collected filaments
    :param controller: Boxmanager_controller obj
    :param filaments: list of filaments
    :param dict_entry_name: filename  file filename in case of SPA, tomo filename in case of folder
    """

    colour = "r"
    colors_ = ["b", "r", "c", "m", "y", "k", "w"]
    is_tomo = controller.model.type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value]

    # i need to fix these 2 internal vars to be able to create the 3D values
    controller.set_has_filament(has_filament=True)
    controller.model.box_dictionary.is_3D = not helper.is_cbox_untraced(b=filaments[0].boxes[0]) if is_tomo and filaments[0] and filaments[0].boxes[0] else False

    for fil_id, fil in enumerate(filaments):
        rects_ = []
        # load the boxes, because the filament could be curvy
        if fil.boxes:

            # in case of multiple import we want to have all the 'new' filament with a given color
            if keep:
                new_rand_color="b"
            else:
                new_rand_color = random.choice(colors_)
                while new_rand_color == colour:
                    new_rand_color = random.choice(colors_)
                colour = new_rand_color
            rects_.extend([box_to_sketch(controller,box, new_rand_color) for box in fil.boxes])
            if not controller.model.box_dictionary.is_3D:
                # load the filament data, we will transform the curvy filaments in straight lines
                box_size = fil.boxes[0].w
                offset = box_size / 2  # for convertion cryolo.BBox to MySketch obj
                filament=helper_filament.picked_filament(box_size=box_size, is_tomo=is_tomo)
                filament.begin_fil = [fil.boxes[0].x + offset, fil.boxes[0].y + offset]
                filament.end_fil = [fil.boxes[-1].x + offset, fil.boxes[-1].y + offset]

                #fill the box_dictionary.filament2D
                z = int(fil.boxes[0].z) if is_tomo else None
                id_fil = controller.model.box_dictionary.add_picked_filament(fil=filament.get_rect_sketch(color=new_rand_color),
                                                                             f=dict_entry_name, n=z,still_picking=False)

                # add color and fil id info in the utils.filament obj (i could use it in case of box resizing)
                fil.meta.update({"color": new_rand_color, "id": id_fil})
                if is_tomo:
                    controller.model.box_dictionary.cryolo_filament_dictionary[dict_entry_name][z].append(fil)
                else:
                    controller.model.box_dictionary.cryolo_filament_dictionary[dict_entry_name].append(fil)
                # fill the box_dictionary.particle2D
                controller.model.box_dictionary.add_sketches(
                    sketches=rects_, f=dict_entry_name,n=z, id_fil=id_fil, colour=new_rand_color)
                controller.model.box_dictionary.update_boxsize_dictionary(new_boxsize=box_size, f=dict_entry_name, n=z)
            else:
                # when we load 3D filament, basically the output of cryolo 'CBOX_FILAMENTS_TRACED', we want to visualize
                # them as list of particles (otherwise we lose information)
                for b in rects_:
                    controller.model.box_dictionary.add_sketches(
                        sketches=[b], f=dict_entry_name, n=int(b.z), id_fil=fil_id, colour=new_rand_color)

    return len(filaments)

def import_SPA(controller:boxmanager_controller, box_dir:str, keep:bool)->bool:
    """
    Import the boxes from files in the SPA case
    In case it cannot (we try to import particle when we are displaying filament or viceversa) return False.
    :param controller: Boxmanager_controller obj
    :param box_dir: path to the folder contains the files
    :param keep: False if we want to discard the data
    """
    t_start = time.time()
    box_imported, filaments_imported = 0, 0
    boxes = list()
    updated_entries = list()
    dict_entry_name = None

    if not path.isdir(box_dir):
        fname=path.splitext(path.basename(box_dir))
        onlyfiles = [fname[0]+fname[1]]
        box_dir = path.dirname(controller.model.box_dir)
    else:
        onlyfiles = [
            f for f in listdir(box_dir)
            if path.isfile(path.join(box_dir, f))
               and not f.startswith(".")
               and path.splitext(f)[0] in helper.get_all_loaded_filesnames(root=controller.view.tree.invisibleRootItem().child(0))
               and (
                       f.endswith(".box")
                       or f.endswith(".txt")
                       or f.endswith(".star")
                       or f.endswith(".cbox")
               )
        ]

    if not onlyfiles:
        return False

    if keep is False:
        rand_color = "r"
    else:
        #self.convert_boxdict_rect_to_list_square()
        rand_color = random.choice(COLORS)
        while rand_color == "r":
            rand_color = random.choice(COLORS)

    pd = controller.view.create_progress_dialog(title="Load box files...")
    pd.show()

    star_dialog_was_shown = False
    is_star_startend = False

    controller.view.is_updating_params = True
    valid_import = False
    for file_index, f in enumerate(onlyfiles):
        pd.setValue(int((file_index + 1) * 100 / len(onlyfiles)))
        controller.view.run_in_same_thread()

        pathfile = path.join(box_dir, f)

        is_helicon = CoordsIO.is_eman1_helicon(path=pathfile)
        is_eman1_startend = CoordsIO.is_eman1_filament_start_end(path=pathfile) if not is_helicon else False
        if not is_helicon and not is_eman1_startend:
            if not star_dialog_was_shown and f.endswith(".star"):
                is_star_startend =  controller.view.qtmessage( msg="Are the star files containing filament coordinates (start/end)?",
                                has_cancel=False) == controller.view.clicked_yes()
                star_dialog_was_shown = True

        # if the file is empty i want just to check the related checkbox
        if os_stat(path.join(box_dir, f)).st_size == 0:
            dict_entry_name = _update_entries_empty_file(f=f, updated_entries=updated_entries, is_star_startend=is_star_startend, is_helicon=is_helicon, is_eman1_startend=is_eman1_startend)
            continue

        if any([is_helicon, is_eman1_startend, is_star_startend]):
            filaments = list()
            dict_entry_name = f[:-4]
            if is_helicon:
                filaments = CoordsIO.read_eman1_helicon(pathfile)
            elif is_eman1_startend:
                filaments = CoordsIO.read_eman1_filament_start_end(pathfile)
            elif is_star_startend:
                print(f"Read filaments {pathfile}")
                filaments = CoordsIO.read_star_filament_file(
                    path=pathfile, box_width=100
                )
                dict_entry_name = f[:-5]
            filaments_imported += _collect_filament(controller,filaments, dict_entry_name, keep)
        else:
            # here it loads the filaments (from .cbox) as list of boxes too
            boxes = _load_data(controller=controller, pathfile=pathfile, keep=keep, t_start=t_start)


        dict_entry_name = path.splitext(f)[0]
        if controller.model.is_cbox and controller.model.has_filament:
            filaments_imported += _collect_filament(controller,boxes,dict_entry_name,keep)
        elif not controller.model.has_filament:
            sketches = [box_to_sketch(controller=controller,box=box, color=rand_color) for box in boxes]
            box_imported += len(sketches)
            controller.model.box_dictionary.add_sketches(sketches=sketches, f=dict_entry_name, n=controller.model.index_tomo, id_fil=None,colour=rand_color)
            # the 'boxsize_dictionary' and 'visualization_dictionary' ll be fill in real time

        updated_entries.append(dict_entry_name)

        # in case we try to load particles and keep the filaments we have an invalid import. We do not change the boxsize.
        # Pay attention: if you try to do that the boxsize will be equal to the height of the last picked/loaded filament
        valid_import = bool(len(boxes)) or valid_import

    # i have to set th boxesize to the GUI, in particle is done into _load_data
    if valid_import and controller.model.has_filament and controller.model.rectangles:
        helper_image.box_size_changed(controller=controller, box_size_from_file=controller.model.rectangles[0].get_width())

    controller.view.set_checkstate_tree_leafs(item=controller.view.tree.invisibleRootItem(), entries=updated_entries, state=controller.view.checkbox_state_value(checked=True))

    print(f"Total time {time.time() - t_start}")
    print(f"Total imported filaments: {filaments_imported}" if controller.model.has_filament else f"Total imported particles: {box_imported}")

    if pd:
        pd.close()

    controller.view.is_updating_params = False

    # set the visualization's options into the visualization combobox
    if controller.model.has_filament:
        type_visualization = constants.Display_visualization_cb.FILAMENT_2D.value
    else:
        type_visualization = constants.Display_visualization_cb.PARTICLE.value
    helper_GUI.set_visualization_combobox(view=controller.view, type_visualization=type_visualization)
    return valid_import


def import_tomo_folder(controller:boxmanager_controller, box_dir:str, keep:bool)->bool:
    """
    Import the boxes from files in the tomo folder case
    In case it cannot (we try to import particle when we are displaying filament or viceversa) return False.
    :param controller: Boxmanager_controller obj
    :param box_dir: path to the folder contains the files
    :param keep: False if we want to discard the data
    """
    controller.model.is_cbox = False
    controller.model.is_cbox_untraced = False
    #self.eman_3D = False
    onlyfiles = [
        f for f in listdir(box_dir)
        if path.isfile(path.join(box_dir, f))
           and not f.startswith(".")
           and path.splitext(f)[0] in helper.get_all_loaded_filesnames(root=controller.view.tree.invisibleRootItem().child(0))
           and (
                   f.endswith(".box")
                   or f.endswith(".star")
                   or f.endswith(".cbox")
           )
    ]

    for index,f in enumerate(onlyfiles):
        if not import_single_tomo(controller=controller, path_3d_file=path.join(box_dir, path.splitext(f)[0]),index_file=index, keep=keep):
            return False
        controller.model.params[path.splitext(f)[0]].box_size = controller.model.boxsize
        controller.view.is_updating_params = True

    # set the boxsize to the current image boxsize
    controller.model.boxsize = controller.model.params[controller.get_current_filename(with_extension=False)].box_size
    controller.view.line_setText(line=controller.view.boxsize_line,value=str(controller.model.boxsize))
    controller.view.is_updating_params = False
    return True

def _setting_boxdictionary_vars(controller:boxmanager_controller)->None:
    """
    After importing boxes I have to set the internal var of Box_dictionary obj
    :param controller: Boxmanager_controller obj
    """
    if controller.model.type_case != constants.Type_case.SPA.value:
        controller.model.box_dictionary.has_filament = controller.model.has_filament
        if controller.model.is_cbox:
            controller.model.box_dictionary.is_cbox = True
            controller.model.box_dictionary.is_3D = True
            controller.model.has_3d = True
        if controller.model.is_cbox_untraced:
            controller.model.has_3d = False
            controller.model.box_dictionary.set_untraced_param()


def import_single_tomo(controller:boxmanager_controller, path_3d_file:str, index_file:Union[int,None], keep:bool)->bool:
    """
    Import the boxes of a single tomo from file.
    In case it cannot (we try to import particle when we are displaying filament or viceversa) return False.

    :param controller: Boxmanager_controller obj
    :param path_3d_file: path to the file
    :param index_file: To avoid to overwrite the checked checkbox into controller.view.set_checkstate_tree_leafs, None if not a folder of tomo
    :param keep: False if we want to discard the data
    """
    filaments_imported,box_imported = 0,0
    t_start = time.time()

    filename=path_3d_file.split("/")[-1]
    pd = controller.view.create_progress_dialog(title="Load box files...")
    pd.show()

    controller.view.is_updating_params = True

    if path.isfile(path_3d_file + ".star") is True:
        boxes = CoordsIO.read_star_file(path=path_3d_file + ".star", box_size=controller.model.boxsize)
    elif path.isfile(path_3d_file + ".box") is True:
        boxes = CoordsIO.read_eman1_boxfile(path=path_3d_file + ".box", is_SPA=False, box_size_default=constants.Default_settings_value.DEFAULT_BOX_SIZE.value)
        #self.eman_3D = True
    elif path.isfile(path_3d_file + ".cbox") is True:
        boxes = CoordsIO.read_cbox_boxfile(path=path_3d_file + ".cbox",min_dist_boxes_filament=0)
        if boxes:
            current_filename = path.splitext(path.basename(path_3d_file + ".cbox"))[0]
            if not _check_correct_import_data(controller=controller,keep=keep, boxes=boxes):
                pd.close()
                return False
            if isinstance(boxes[0], utils.Filament):
                filaments_imported += _collect_filament(controller=controller,filaments=boxes,dict_entry_name=current_filename,keep=keep)
            else:
                controller.set_has_filament(has_filament=False)
            if boxes[0]:
                b = boxes[0].boxes[0] if controller.model.has_filament else boxes[0]
                vis = constants.Visualization_cb.CIRCLE_SEGMENTED.value if controller.model.has_filament else constants.Visualization_cb.CIRCLE.value
                controller.model.is_cbox = not helper.is_cbox_untraced(b=b)
                controller.model.has_3d = controller.model.is_cbox
                controller.model.is_cbox_untraced = helper.is_cbox_untraced(b=b)
                sketch = box_to_sketch(controller=controller, box=b, color=None)
                controller.model.est_box_from_cbox = sketch.get_est_size(circle=False)
                box_size =sketch.get_width()
                helper_image.box_size_changed(controller=controller, box_size_from_file=box_size)
                for z in controller.model.box_dictionary.boxsize_dictionary[controller.get_current_filename(with_extension=False)]:
                    controller.model.box_dictionary.update_boxsize_dictionary(new_boxsize=box_size,f=current_filename,n=z)
                    controller.model.box_dictionary.update_visualization_dictionary(new_visualization=vis, f=current_filename,n=z)
    else:
        controller.view.err_message(msg=f"Error: '{path_3d_file}' is not valid .star, .box or .cbox files")
        pd.close()
        return False

    # set cbox and filament vars into the model.box_dictionary var
    if controller.model.type_case != constants.Type_case.SPA.value:
        _setting_boxdictionary_vars(controller=controller)

    if keep is False:
        rand_color = "r"
    else:
        rand_color = random.choice(["b", "c", "m", "y", "k", "w"])

    if not controller.model.has_filament :
        for box in boxes:
            sketch = box_to_sketch(controller=controller, box=box, color=rand_color)
            box_imported += 1
            controller.model.box_dictionary.add_sketch(sketch=sketch, f=filename, n=int(box.z))

    updated_entries = [str(int(z)) for z in CoordsIO.read_cbox_include_list(path=path_3d_file + ".cbox")]
    item =  controller.view.tree.invisibleRootItem().child(0).child(index_file) if index_file is not None else controller.view.tree.invisibleRootItem()
    controller.view.set_checkstate_tree_leafs(item=item, entries=updated_entries, state=controller.view.checkbox_state_value(checked=True))

    print(f"Total time {time.time() - t_start}")
    print(f"Total imported filaments: {filaments_imported}" if controller.model.has_filament else f"Total imported particles: {box_imported}")

    pd.close()
    controller.view.is_updating_params = False

    # set the visualization's options into the visualization combobox
    type_visualization = constants.Display_visualization_cb.FILAMENT_2D.value
    if not controller.model.has_filament:
        type_visualization = constants.Display_visualization_cb.PARTICLE.value
    elif controller.model.has_3d:
        type_visualization = constants.Display_visualization_cb.FILAMENT_3D.value
    helper_GUI.set_visualization_combobox(view=controller.view, type_visualization=type_visualization)

    return True

def import_data(controller:boxmanager_controller)->None:
    """
    :param controller: Boxmanager_controller obj
    """
    if not controller.model.box_dictionary or (controller.model.unsaved_changes and controller.view.qtmessage(
                                msg="There are unsaved changes. Are you sure?", has_cancel=True) == controller.view.clicked_cancel()):
        return

    # False when we want to discard the boxes already present in boxdictionary
    keep = False

    if not controller.model.box_dictionary.is_empty():
        if controller.view.qtmessage(msg= "Keep old boxes loaded and show the new ones in a different color?",
                                has_cancel=False) == controller.view.clicked_yes():
            keep = True
        else:
            pass
            #self.has_filament = False
            #self.is_cbox_untraced = False


    # i do not need if self.plot is not None or self.is_folder_3D_tomo is True:
    box_dir = controller.view.qt_esixting_folder(msg="Select Box Directory")

    if not box_dir:
        return

    if not keep:
        _remove_old_boxes(controller=controller)

    if controller.model.type_case == constants.Type_case.SPA.value:
        valid_import = import_SPA(controller=controller, box_dir=box_dir, keep=keep)
    elif controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
        valid_import = import_tomo_folder(controller=controller, box_dir=box_dir, keep=keep)
    else:
        valid_import= import_single_tomo(controller=controller, path_3d_file=path.join(box_dir, controller.get_current_filename(with_extension=False)), index_file=None, keep=keep)

    if not valid_import:
        return

    if keep and controller.model.has_filament:
        controller.model.box_dictionary.set_colors_all(new_color="w")

    kind_of_cbox =controller.model.is_cbox or controller.model.is_cbox_untraced
    controller.view.thresholding_tab_blurring(is_enable=kind_of_cbox, kind_of_cbox=kind_of_cbox, has_filament=controller.model.has_filament,type_case=controller.model.type_case)
    controller.view.tracing_tab_blurring(is_enable=controller.model.is_cbox_untraced,has_filament=controller.model.has_filament)

    controller.view.blur_picking_combobox(has_filament=controller.model.has_filament, is_importing=True)

    if kind_of_cbox or controller.model.box_dictionary.has_filament:
        controller.view.set_cbox_visualization(has_filament=controller.model.has_filament)
    # load the boxes present in the current image/slice, boxsize and visualization get from GUI (already saved in the model)
    controller.model.rectangles = controller.model.box_dictionary.get_sketches(
                                f=controller.get_current_filename(with_extension=False), n=controller.model.index_tomo, is_3D=controller.model.has_3d, is_tracing = False)

    # restore the canvas
    if controller.view.background_current:
        controller.view.fig.canvas.restore_region(controller.view.background_current)

    # draw the sketches present in the current image/slice
    helper_image.draw_all_patches(controller=controller)
    controller.view.fig.canvas.blit(controller.view.ax.bbox)

    # update counters
    controller.view.update_tree_boxsizes( update_current=False,all_tomos=True)
    controller.view.update_all_global_counters()
    # in case of cbox filament some options are not available
    if controller.model.has_filament:
        controller.view.blur_estimated_size()
        controller.model.box_distance_filament = controller.model.box_dictionary.set_box_distance_dict_after_import()
        controller.view.line_setText(line=controller.view.box_distance_filament_line, value=str(controller.model.box_distance_filament))


"""
In this file there are some functions to manage the GUI and the pyQT import
"""

try:
    QT = 4
    import PyQt4.QtGui as QtG
    import PyQt4.QtCore as QtCore
    from PyQt4.QtGui import QFontMetrics,QFont
    import matplotlib.backends.backend_qt4agg as plt_qtbackend
except ImportError:
    QT = 5
    import PyQt5.QtWidgets as QtG
    from PyQt5.QtGui import QFontMetrics,QFont
    import PyQt5.QtCore as QtCore
    import matplotlib.backends.backend_qt5agg as plt_qtbackend

from cryoloBM import helper,constants
import numpy as np
from os import path
from typing import List

from cryoloBM import boxmanager_view,boxmanager_controller    # for type hinting purpose


def apply_to_all_the_tomo_question(view:boxmanager_view, name_param:str)->int:
    msg = f"Do you want to change '{ name_param} ' to all tomographies"
    return QtG.QMessageBox.question(view, "Message", msg, QtG.QMessageBox.Yes, QtG.QMessageBox.No)


def err_message(view:boxmanager_view, msg:str)->None:
    """
    shows the pop up window with the 'msg' error message
    :param view: Boxmanager_view obj
    :param msg: message visualized on the pop up window
    """
    errmsg = QtG.QErrorMessage(view)
    errmsg.showMessage(msg)

def qtmessage(view:boxmanager_view, msg:str, has_cancel:bool)->int:
    """
    shows the pop up question message and returns the inserted answer
    :param view: Boxmanager_view obj
    :param msg: message visualized on the pop up window
    :param has_cancel: If True it has 'QtG.QMessageBox.Cancel' instead of 'QtG.QMessageBox.No'
    :return 'QtG.QMessageBox.question'
    """
    second_choice = QtG.QMessageBox.Cancel if has_cancel else QtG.QMessageBox.No
    return QtG.QMessageBox.question(view, "Message", msg, QtG.QMessageBox.Yes, second_choice)

def qt_esixting_folder(view:boxmanager_view, msg:str)->str:
    """
    shows the pop up question message and returns the inserted answer
    :param view: Boxmanager_view obj
    :param msg: message visualized on the pop up window
    :return the path as string
    """
    return str( QtG.QFileDialog.getExistingDirectory(view, msg) )

def get_selected_folder(view:boxmanager_view, unsaved_changes:bool)->str:
    """
    Set image_folder_path. It is selected by the user via GUI
    :param view: Boxmanager_view obj
    :param unsaved_changes:
    :return The selected folder
    """
    selected_folder = qt_esixting_folder(view, msg = "Select Image Directory")

    if selected_folder == "" or (
            unsaved_changes and qtmessage(view, msg="All loaded boxes are discarded. Are you sure?",
                                                has_cancel=True) == QtG.QMessageBox.Cancel):
        return ""
    return  selected_folder

def get_selected_file(view:boxmanager_view, unsaved_changes:bool)->str:
    """
    Set current_image_path. It is selected by the user via GUI
    :param view: Boxmanager_view obj
    :param unsaved_changes:
    :return The selected file
    """
    selected_image = QtG.QFileDialog.getOpenFileName(view, "Select 3D Image File")[0]

    if selected_image == "" or (
            unsaved_changes and qtmessage(view, msg="All loaded boxes are discarded. Are you sure?",
                                                has_cancel=True) == QtG.QMessageBox.Cancel):
        return ""
    return selected_image

def add_item_combobox(name:str, obj:QtG.QComboBox)->None:
    """
    Adds the 'name' item of the 'obj' combobox if it is not present
    :param name: string name of the item
    :param obj: QtG.QComboBox obj
    """
    pos = get_pos_item_combobox(name=name, obj=obj)
    if pos == -1:
        obj.addItems([name])


def remove_item_combobox(name:str, obj:QtG.QComboBox)->None:
    """
    Removes the 'name' item of the 'obj' combobox if it is present
    :param name: string name of the item
    :param obj: QtG.QComboBox obj
    """
    pos = get_pos_item_combobox(name=name, obj=obj)
    if pos > -1:
        obj.removeItem(pos)

def get_pos_item_combobox(name:str, obj:QtG.QComboBox)->int:
    """
    Returns the position of the 'name' item of the 'obj' combobox. -1 f it is not present
    :param name: string name of the item
    :param obj: QtG.QComboBox obj
    :return: position in list
    """
    #AllItems = [ for i in range(obj.count())]
    for i in range(obj.count()):
        if name == obj.itemText(i):
            return i
    return -1

def delete_all_items_combobox(obj:QtG.QComboBox)->None:
    """
    Deletes all the items in the given combobox obj
    :param obj: QtG.QComboBox obj
    """
    for i in range(obj.count()-1,-1,-1):
        remove_item_combobox(obj.itemText(i),obj)

def fill_root_childs(root:QtG.QTreeWidgetItem, tot_slices:int, is_folder:bool =False, root_child_index:int =None)->None:
    """
    Creates all the checkboxes of a given tomogram in the GUI
    :param root: The 'QTreeWidgetItem' obj represent the item list (namely the filenames)
    :param tot_slices:  number of slices present in the tomogram
    :param is_folder: True if we loaded a folder of tomo
    :param root_child_index: index of child in root. It is the item, which represent a tomogram in a folder
    """
    for i in range(tot_slices):
        QtCore.QCoreApplication.instance().processEvents()
        c = QtG.QTreeWidgetItem([str(i)])
        c.setCheckState(0, QtCore.Qt.Unchecked)
        if is_folder:
            root.child(root_child_index).addChild(c)
        else:
            root.addChild(c)

def uncheck_all_slides(view:boxmanager_view, type_case:int)->None:
    """
    Uncheck all the checked checkboxes
    :param view: Boxmanager_view obj
    :param type_case: constants.Type_case. value
    """
    # uncheck all the checked slices (tomo cases) or images (SPA cases)
    for root_index in range(view.tree.invisibleRootItem().childCount()):
        root_element = view.tree.invisibleRootItem().child(root_index)  # can be tomogram or a folder
        root_element.setCheckState(0, QtCore.Qt.Unchecked)
        for child_index in range(root_element.childCount()):
            root_element.child(child_index).setCheckState(0, QtCore.Qt.Unchecked)
            if type_case == constants.Type_case.TOMO_FOLDER.value:
                for child_child_index in range(root_element.child(child_index).childCount()):
                    root_element.child(child_index).child(child_child_index).setCheckState(0, QtCore.Qt.Unchecked)

def event_checkbox_changed(item:QtG.QTreeWidgetItem, column:int)->None:
    """
    Clicking on a node of a checkbox tree set as checked all its children recursively
    It is called directly by box_manager_view. It fills the input params in automatic
    """
    if column == 0:
        if item.childCount()>0:
            for child_index in range(item.childCount()):
                item.child(child_index).setCheckState(0,item.checkState(0))

def tracing_tab_blurring(view:boxmanager_view, is_enable:bool, has_filament:bool)->None:
    """
    After loading from file we have to unblur some options in tracing  tabs
    :param view: Boxmanager_view obj
    :param is_enable: If True enable the options
    :param has_filament: If True we work with filament
    """

    view.search_range_label.setEnabled(is_enable)
    view.search_range_line.setEnabled(is_enable)
    view.search_range_slider.setEnabled(is_enable)
    view.memory_line.setEnabled(is_enable)
    view.memory_label.setEnabled(is_enable)
    view.memory_slider.setEnabled(is_enable)
    view.min_length_label.setEnabled(is_enable)
    view.min_length_line.setEnabled(is_enable)
    view.min_length_slider.setEnabled(is_enable)
    view.preview_label.setEnabled(is_enable)
    view.preview_checkbox.setEnabled(is_enable)
    view.button_trace.setEnabled(is_enable)

    # usable only in filament case
    is_enable = is_enable and has_filament  # True only if both are True

    view.min_edge_weight_slider.setEnabled(is_enable)
    view.min_edge_weight_line.setEnabled(is_enable)
    view.min_edge_weight_label.setEnabled(is_enable)
    view.win_size_line.setEnabled(is_enable)
    view.win_size_slider.setEnabled(is_enable)
    view.win_size_label.setEnabled(is_enable)

def thresholding_tab_blurring_(view:boxmanager_view, is_enable:bool, kind_of_cbox:bool, has_filament:bool, type_case:int)->None:
    """
    After loading from file we have to unblur some options in thresholding tabs
    :param view: Boxmanager_view obj
    :param is_enable: If True enable SOME options. If False disable ALL the options
    :param kind_of_cbox: True if model.is_cbox or model.is_cbox_untracved is True
    :param has_filament: If True we work with filament
    :param type_case: constants.Type_case. value
    """
    if kind_of_cbox or not is_enable:
        view.conf_thresh_line.setEnabled(is_enable)
        view.conf_thresh_slide.setEnabled(is_enable)
        view.conf_thresh_label.setEnabled(is_enable)
        view.use_estimated_size_label.setEnabled(is_enable)
        view.use_estimated_size_checkbox.setEnabled(is_enable)
        view.show_confidence_histogram_action.setEnabled(is_enable)
        view.show_size_distribution_action.setEnabled(is_enable)
        if not has_filament or not is_enable:
            view.upper_size_thresh_label.setEnabled(is_enable)
            view.upper_size_thresh_slide.setEnabled(is_enable)
            view.upper_size_thresh_line.setEnabled(is_enable)
            view.lower_size_thresh_label.setEnabled(is_enable)
            view.lower_size_thresh_slide.setEnabled(is_enable)
            view.lower_size_thresh_line.setEnabled(is_enable)
        if type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value] or not is_enable:
            view.num_boxes_thresh_slide.setEnabled(is_enable)
            view.num_boxes_thresh_line.setEnabled(is_enable)
            view.num_boxes_thres_label.setEnabled(is_enable)

def update_low_up_thresh(view:boxmanager_view,max_size:int,min_size:int)->None:
    """
    After loading from file we have to update upper and lower threshold values
    :param view: Boxmanager_view obj
    :param max_size: max size value
    :param min_size: min size value
    """
    view.is_updating_params = True
    view.upper_size_thresh_slide.setMaximum(max_size)
    view.upper_size_thresh_slide.setMinimum(min_size)
    view.upper_size_thresh_slide.setValue(max_size)
    view.upper_size_thresh_line.setText("" + str(max_size))
    view.lower_size_thresh_slide.setMaximum(max_size)
    view.lower_size_thresh_slide.setMinimum(min_size)
    view.lower_size_thresh_slide.setValue(min_size)
    view.lower_size_thresh_line.setText("" + str(min_size))
    view.is_updating_params = False

def set_checkstate_tree_leafs( item:QtG.QTreeWidgetItem, entries:List[str], state:QtCore.Qt.CheckState)->None:
    """
    Check or uncheck the checkboxes in the tree
    :param item: controller.view.tree.invisibleRootItem()
    :param entries: list of entries as strings (e.g.: list of slices)
    :param state: helper_GUI.QtCore.Qt.Checked or helper_GUI.QtCore.Qt.Unchecked
    """
    child_count = item.childCount()
    child_child_counter = item.child(0).childCount()
    if child_child_counter>0:
        for child_id in range(child_count):
            set_checkstate_tree_leafs(item.child(child_id),entries,state)
    else:
        for child_id in range(child_count):
            parent_identifier = path.splitext(item.text(0))[0]
            dict_identifier = path.splitext(item.child(child_id).text(0))[0]
            if len(entries) == 0:
                return
            if type(entries[0]) is tuple:
                parent_entries = [entry[0] for entry in entries]
                child_entries = [entry[1] for entry in entries]
                for k in range(len(entries)):
                    if parent_identifier == parent_entries[k] and dict_identifier == child_entries[k]:
                        item.child(child_id).setCheckState(0,state)

            else:
                if dict_identifier in entries:
                    item.child(child_id).setCheckState(0,state)

def display_default_params(view:boxmanager_view)->None:
    """
    After resetting (file->reset or loading new images) i have to set the default vars as at the start
    :param view: boxmanager_view obj
    """
    view.is_updating_params = True
    view.upper_size_thresh_line.setText(str(constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value))
    view.lower_size_thresh_line.setText(str(constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value))
    view.conf_thresh_line.setText(str(constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value))
    view.num_boxes_thresh_line.setText(str(constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value))
    view.memory_line.setText(str(constants.Default_settings_value.DEFAULT_MEMORY.value))
    view.search_range_line.setText(str(constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value))

    view.min_length_line.setText(str(constants.Default_settings_value.DEFAULT_MIN_LENGTH.value))
    view.win_size_line.setText(str(constants.Default_settings_value.DEFAULT_BOX_SIZE.value))
    view.min_edge_weight_line.setText(str(constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value))
    view.is_updating_params = False


def set_visibility_box_distance(view,is_visible: bool) -> None:
        """
        It makes the 'box distance' buttons and line of the 'visualization' tab visible or not
        :param view: boxmanager_view obj
        :param is_visible: True makes visible these widgets
        """
        view.button_set_box_distance_filament.setVisible(is_visible)
        view.box_distance_filament_label.setVisible(is_visible)
        view.box_distance_filament_line.setVisible(is_visible)

def set_visualization_combobox(view, type_visualization=0) -> None:
    """
    It fills the 'visualization' combobox of the 'visualization' tab in function of the picking combobox value
    :param view: boxmanager_view obj
    :param type_visualization: see constant.Display_visualization_cb.value
    """
    if not view.is_updating_params:
        delete_all_items_combobox(obj=view.visualization_combobox)
        if type_visualization == constants.Display_visualization_cb.PARTICLE.value:
            values = [constants.Visualization_cb.RECT.value, constants.Visualization_cb.CIRCLE.value]
        elif type_visualization == constants.Display_visualization_cb.FILAMENT_2D.value:
            values = [constants.Visualization_cb.RECT_FILAMENT_START_END.value,
                  constants.Visualization_cb.RECT_FILAMENT_SEGMENTED.value,
                  constants.Visualization_cb.CIRCLE_SEGMENTED.value]
        elif type_visualization == constants.Display_visualization_cb.FILAMENT_3D.value:
            values = [constants.Visualization_cb.RECT_FILAMENT_SEGMENTED.value,
                      constants.Visualization_cb.CIRCLE_SEGMENTED.value]
        else:
            values=list()
            print("WARNING: invalid 'constants.Display_visualization_cb' value")
        for item_name in values:
            add_item_combobox(name=item_name, obj=view.visualization_combobox)

def set_picking_combobox(view, is_enabled: bool) -> None:
    """
    It fills the picking combobox of the main window
    :param view: boxmanager_view obj
    :param is_enabled: False it is blurred
    """
    values = [constants.Picking_cb.PARTICLE.value, constants.Picking_cb.FILAMENT.value]

    delete_all_items_combobox(obj=view.picking_filament_combox)
    for item_name in values:
        add_item_combobox(name=item_name, obj=view.picking_filament_combox)

    view.picking_filament_combox.setEnabled(is_enabled)


from cryoloBM import helper_image,helper_GUI,boxmanager_controller

try:
    QT = 4
    import PyQt4
    from matplotlib.backends.backend_qt4agg import (
        NavigationToolbar2QT as NavigationToolbar,
    )
except ImportError:
    QT = 5
    import PyQt5
    from matplotlib.backends.backend_qt5agg import (
        NavigationToolbar2QT as NavigationToolbar,
    )


class Boxmanager_Toolbar(NavigationToolbar):
    def __init__(self,  controller:boxmanager_controller)->None:
        self.controller=controller
        self.dozoom = False
        NavigationToolbar.__init__(self, self.controller.view.plot.canvas, self.controller.view.plot)
        self.background_origin = None
        self.start_point_zoom = None

    def press_zoom(self, event:helper_GUI.plt_qtbackend)->None:
        self.start_point_zoom = [event.xdata ,event.ydata ]
        super(Boxmanager_Toolbar, self).press_zoom(event)


    def zoom(self, *args)->None:
        # the first time i click the zoom i save the whole image
        if not self.background_origin:
            self.background_origin = self.controller.view.background_current
        super(Boxmanager_Toolbar, self).zoom(args)

    def home(self, *args)->None:
        helper_image.delete_all_patches(controller=self.controller)
        if self.background_origin:
            self.controller.view.fig.canvas.restore_region(self.background_origin)
        self.controller.view.background_current = self.controller.view.fig.canvas.copy_from_bbox(
            self.controller.view.ax.bbox)
        self.controller.view.background_current_onmove = self.controller.view.background_current

        self.controller.model.rectangles = self.controller.model.box_dictionary.get_sketches(
            f=self.controller.get_current_filename(with_extension=False),
            n=self.controller.model.index_tomo, is_3D=self.controller.model.has_3d,  is_tracing = False)
        helper_image.draw_all_patches(controller=self.controller)
        super(Boxmanager_Toolbar, self).home(args)

    def release_zoom(self, event:helper_GUI.plt_qtbackend)->None:
        self.dozoom = False

        lastx = self.start_point_zoom[0]
        lasty = self.start_point_zoom[1]
        # ignore singular clicks - 5 pixels is a threshold
        if not (abs(event.x - lastx) < 5 or abs(event.y - lasty) < 5):
            self.dozoom = True
            helper_image.delete_all_patches(controller=self.controller)
            self.controller.view.fig.canvas.draw()
            self.controller.view.background_current = self.controller.view.fig.canvas.copy_from_bbox(self.controller.view.ax.bbox)
            self.controller.view.background_current_onmove = self.controller.view.background_current
            self.controller.view.zoom_update = True

        super(Boxmanager_Toolbar, self).release_zoom(event)
        self.start_point_zoom = None


    def pan(self, *args)->None:
        super(Boxmanager_Toolbar, self).pan(args)

    def drag_pan(self, event:helper_GUI.plt_qtbackend)->None:
        super(Boxmanager_Toolbar, self).drag_pan(event)
        helper_image.delete_all_patches(controller=self.controller)
        if self.controller.view.background_current:
            self.controller.view.fig.canvas.restore_region(self.controller.view.background_current)
        self.controller.view.background_current = self.controller.view.fig.canvas.copy_from_bbox(self.controller.view.ax.bbox)
        self.controller.view.zoom_update = True



"""
The following functions manage the changes due to the modifications of the values of the table thresholding
"""

#todo the 'number boxes threshold' is not implemented

from cryoloBM import constants, helper_image, helper_tab_tracing
from cryoloBM import boxmanager_controller # for type hinting purpose

def update_after_setting(controller:boxmanager_controller,update_all:bool)->None:
    """
    Manage the operation trigged by changing a threshold param:
    -) delete old sketches on the current image
    -) draw new sketches on the current image
    -) update the counters
    :param controller: Boxmanager_controller obj
    :param update_all: True if we have to update the counters for all the tomos (only TOMO FOLDER cas)
    """

    # i do not need to call h_keypress() but only reset its related flag
    if controller.model.toggle:
        controller.model.toggle = False

    # udpate the current image visualization
    if controller.model.rectangles:
        #remove the old patches, they are already saved in controller.model.rectangles
        helper_image.delete_all_patches(controller=controller)

        # restore the canvas
        controller.view.fig.canvas.restore_region(controller.view.background_current)

        # redraw the sketches. this function will filter by threshold values
        helper_image.draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)

    # update the counters
    controller.view.update_tree_boxsizes(update_current=False, all_tomos=update_all)
    if update_all:
        controller.view.update_all_global_counters()
    else:
        controller.view.update_global_counter()

def conf_thresh_changed(controller:boxmanager_controller)->None:
    """
    Run when the confidence threshold is changed
    :param controller: Boxmanager_controller obj
    """
    # self.trace_all = False
    update_all = None
    try:
        controller.model.current_conf_thresh = float(controller.view.conf_thresh_slide.value()) / 100
        if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
            update_all =controller.view.apply_to_all_the_tomo_question( name_param = "confidence threshold") == controller.view.clicked_yes()
            if update_all:
                #self.trace_all = True
                for p in controller.model.params.values():
                    p.conf_thresh = controller.model.current_conf_thresh

            else:
                controller.model.params[controller.get_current_filename(with_extension=False)].conf_thresh = controller.model.current_conf_thresh
    except ValueError:
        return

    controller.view.line_setText(line=controller.view.conf_thresh_line, value=str(controller.model.current_conf_thresh))
    update_after_setting(controller=controller,update_all=update_all)
    controller.view.is_updating_params = True

def conf_thresh_label_changed(controller:boxmanager_controller)->None:
    """
    Run when the label of the confidence threshold is changed
    :param controller: Boxmanager_controller obj
    """
    try:
        old_value = controller.model.current_conf_thresh
        new_value = float(controller.view.conf_thresh_line.text())
        if new_value > 1.0 or new_value < 0:
            if not controller.view.is_updating_params:
                controller.view.err_message(msg=f"value has to be between 0 and 1")
                controller.view.line_setText(line=controller.view.conf_thresh_line,value=str(old_value))
            return
    except ValueError:
        return

    controller.model.current_conf_thresh = new_value
    controller.view.conf_thresh_slide.setValue(new_value * 100)
    if controller.view.preview_is_on:
        helper_tab_tracing.display_preview(controller=controller,only_reload=True)

def lower_size_thresh_changed(controller:boxmanager_controller)->None:
    """
    Run when the lower size threshold is changed
    :param controller: Boxmanager_controller obj
    """
    #self.trace_all = False
    update_all = False
    controller.model.lower_size_thresh = int(float(controller.view.lower_size_thresh_slide.value()))
    if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
        update_all =controller.view.apply_to_all_the_tomo_question( name_param = "lower size threshold") == controller.view.clicked_yes()
        # self.trace_all = True
        if update_all:
            for p in controller.model.params.values():
                p.lower_size_thresh = controller.model.lower_size_thresh
        else:
            controller.model.params[controller.get_current_filename(with_extension=False)].lower_size_thresh = controller.model.lower_size_thresh

    controller.view.line_setText(line=controller.view.lower_size_thresh_line, value=str(controller.model.lower_size_thresh))
    update_after_setting(controller=controller, update_all=update_all)
    controller.view.is_updating_params = True


def lower_size_label_changed(controller:boxmanager_controller)->None:
    """
    Run when the label of the lower size threshold is changed
    :param controller: Boxmanager_controller obj
    """
    try:
        new_value = int(float(controller.view.lower_size_thresh_line.text()))
        old_value = controller.model.lower_size_thresh
        if new_value > controller.model.max_valid_size_thresh or new_value < controller.model.min_valid_size_thresh:
            if not controller.view.is_updating_params:
                controller.view.err_message(msg=f"value has to be between { controller.model.min_valid_size_thresh} and { controller.model.max_valid_size_thresh}")
            controller.view.line_setText(line=controller.view.lower_size_thresh_line,value=str(old_value))
            return
    except ValueError:
        return

    controller.view.lower_size_thresh_slide.setValue(new_value)
    controller.model.lower_size_thresh = new_value
    if controller.view.preview_is_on:
        helper_tab_tracing.display_preview(controller=controller,only_reload=True)

def upper_size_thresh_changed(controller:boxmanager_controller)->None:
    """
    Run when the upper size threshold is changed
    :param controller: Boxmanager_controller obj
    """
    # self.trace_all = False
    update_all = False
    controller.model.upper_size_thresh = int(float(controller.view.upper_size_thresh_slide.value()))
    if not controller.view.is_updating_params and controller.model.type_case == constants.Type_case.TOMO_FOLDER.value and not controller.view.is_slider_pressed:
        update_all = controller.view.apply_to_all_the_tomo_question(name_param="upper size threshold") == controller.view.clicked_yes()
        # self.trace_all = True
        if update_all:
            for p in controller.model.params.values():
                p.upper_size_thresh = controller.model.upper_size_thresh
        else:
            controller.model.params[controller.get_current_filename(with_extension=False)].upper_size_thresh = controller.model.upper_size_thresh

    controller.view.line_setText(line=controller.view.upper_size_thresh_line, value=str(controller.model.upper_size_thresh))
    update_after_setting(controller=controller, update_all=update_all)
    controller.view.is_updating_params = True

def upper_size_label_changed(controller:boxmanager_controller)->None:
    """
    Run when the label of the upper size threshold is changed
    :param controller: Boxmanager_controller obj
    """
    try:
        new_value = int(float(controller.view.upper_size_thresh_line.text()))
        old_value = controller.model.upper_size_thresh
        if new_value > controller.model.max_valid_size_thresh or new_value < controller.model.min_valid_size_thresh:
            if not controller.view.is_updating_params:
                controller.view.err_message(msg= f"value has to be between {controller.model.min_valid_size_thresh} and {controller.model.max_valid_size_thresh}")
            controller.view.line_setText(line=controller.view.upper_size_thresh_line,value=str(old_value))
            return
    except ValueError:
        return

    controller.view.upper_size_thresh_slide.setValue(new_value)
    controller.model.upper_size_thresh = new_value
    if controller.view.preview_is_on:
        helper_tab_tracing.display_preview(controller=controller,only_reload=True)


import glob
import sys
import os
import numpy as np
import csv
import copy
import re
from cryolo import CoordsIO, utils
from cryoloBM.bmtool import BMTool
import cryoloBM.boxmanager_tools as btools
from argparse import ArgumentParser
import argparse

class ScaleTool(BMTool):

    def get_command_name(self) -> str:
        return "scale"

    def create_parser(self, parentparser : ArgumentParser) -> ArgumentParser:
        parser_scale = parentparser.add_parser(
            self.get_command_name(),
            help="Scales all sort of coordinate files.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        scale_required_group = parser_scale.add_argument_group(
            "Required arguments",
            "Scales coordinate files.",
        )

        scale_required_group.add_argument(
            "-i",
            "--input",
            required=True,
            help="Input folder or file.",
        )

        scale_required_group.add_argument(
            "-o",
            "--output",
            required=True,
            help="Output folder where to write the scaled coordinates.",
        )

        scale_required_group.add_argument(
            "-s",
            "--scale",
            type=float,
            required=True,
            help="Scale factor",
        )

        return parser_scale

    def run(self, args):
        scale(
            input_path=args.input,
            output_path=args.output,
            scale_factor=args.scale
        )



def read_eman1_helicon_micrgraph_name(path):
    with open(path, "r") as csvfile:
        csvlines = csvfile.readlines()
        for index, row in enumerate(csvlines):
            if row.startswith("#micrograph"):
                pat = "#micrograph:\s(.+)"
                reexp_res = re.findall(pat,row)
                return reexp_res[0]

def write_coords_file(path, array):
    np.savetxt(path, array, delimiter=" ", fmt="%10.5f")

    '''
    with open(path, "w") as coordsfile:
        boxwriter = csv.writer(
            coordsfile, delimiter=" ", quotechar="|", quoting=csv.QUOTE_NONE
        )

        for row in array:
            boxwriter.writerow([row[0], row[1], row[2]])
    '''


def scale(input_path,
          output_path,
          scale_factor):
    '''
    This function reads the coordinates in the input_type given the input_type, scales them
    according the scale_factor and write them back into the output_path
    :param input_path:
    :param output_path:
    :param scale_factor:
    :return:
    '''

    # 1. Read data according input_type.
    # 2. Call respective scale function
    # 3. Write to disk

    os.makedirs(output_path, exist_ok=True)
    if os.path.isfile(input_path):
        files = [input_path]
    else:
        path = os.path.join(os.path.abspath(input_path), "*")
        files = glob.glob(path)

    for pth in files:
        if btools.get_file_type(pth) == btools.TYPE_COORDS:
            coords = [np.atleast_2d(np.genfromtxt(pth))]
            scaled_coords = scale_coords(coords, scale_factor)
            output_file_path = os.path.join(output_path,os.path.basename(pth))
            write_coords_file(output_file_path,scaled_coords[0])
        elif btools.get_file_type(pth) == btools.TYPE_EMAN_HELICON:
            filaments_per_file = CoordsIO.read_eman1_helicon(pth)
            mic_name = read_eman1_helicon_micrgraph_name(pth)
            scaled_filaments = scale_filament(filaments_per_file,scale_factor)
            output_file_path = os.path.join(output_path, os.path.basename(pth))
            CoordsIO.write_eman1_helicon(scaled_filaments, output_file_path, mic_name)
        elif btools.get_file_type(pth) == btools.TYPE_CBOX:
            coords = CoordsIO.read_cbox_boxfile(pth)
            addinclude = CoordsIO.read_cbox_include_list(pth)
            scaled_coords = []
            for elem in coords:
                if isinstance(elem,utils.Filament):
                    scaled_elem = scale_filament([elem], scale_factor)[0]
                else:
                    scaled_elem = scale_boxes([elem],scale_factor)[0]
                scaled_coords.append(scaled_elem)
            output_file_path = os.path.join(output_path, os.path.basename(pth))
            CoordsIO.write_cbox_file(path=output_file_path, coordinates=scaled_coords,additional_slice_include=addinclude)


        else:
            print("File type not supported yet", pth)

def scale_boxes(boxes, scale):
    scaled_boxes = []
    for box in boxes:
        sbox = copy.deepcopy(box)
        sbox.x = sbox.x * scale
        sbox.y = sbox.y * scale
        if sbox.z is not None:
            sbox.z = sbox.z * scale
        sbox.h = sbox.h * scale
        sbox.w = sbox.w * scale
        scaled_boxes.append(sbox)
    return scaled_boxes

def scale_filament(filaments, scale):
    scaled_filaments = []
    for fil in filaments:
        sfil = copy.deepcopy(fil)
        sfil.boxes = scale_boxes(sfil.boxes,scale)
        scaled_filaments.append(sfil)

    return scaled_filaments

def scale_coords(list_of_coords, scale):
    scaled = []
    for coords in list_of_coords:
        coords[:,:3] = coords[:,:3] * scale
        scaled.append(coords)
    return scaled


def scale_fil_seg():
    pass



from typing import List, Dict
from cryolo.utils import Filament, resample_filaments, BoundBox
from cryoloBM.bmtool import BMTool
from argparse import ArgumentParser
import argparse
import numpy as np
import os

class FilamentResampler():

    def __init__(
            self,
            box_distance: float,
    ):
        self.box_distance = box_distance

    def resampling(self, filaments : List[Filament]) -> List[Filament]:
        """

        :param filaments: List of filaments
        :param boxdistance: box distance
        :return: List of resampled filaments
        """
        return resample_filaments(filaments, self.box_distance)

class FilamentResampleTool(BMTool):

    def get_command_name(self) -> str:
        return "resample"

    def create_parser(self, parser=None) -> ArgumentParser:

        if parser is None:
            filament_parser = ArgumentParser(
                self.get_command_name(),
                description="Resamples 3D filaments coords (.coords format) to different box distance. "
                     "This tool is provisional and might disappear in future versions.",
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )
        else:
            filament_parser = parser.add_parser(
                self.get_command_name(),
                help="Resamples 3D filaments coords (.coords format) to different box distance. "
                     "This tool is provisional and might disappear in future versions.",
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )

        req_group = filament_parser.add_argument_group(
            "Required arguments",
            "Resamples filaments coords to different box distance",
        )

        req_group.add_argument(
            "-i",
            "--input",
            required=True,
            help="Path to folder with .coords input files. Only those with FID can be used as input.",
        )
        req_group.add_argument(
            "-d",
            "--distance",
            required=True,
            type=float,
            help="New box distance.",
        )

        req_group.add_argument(
            "-o",
            "--output",
            required=True,
            help="Output folder where to write the resampled filaments. ",
        )

        return filament_parser

    def coords_to_filaments(self, data : np.array) -> List[Filament]:
        fils = []
        for id in np.unique(data[:,3]):
            boxes = []
            for row in data[data[:,3]==id]:
                x = row[0]
                y = row[1]
                z = row[2]
                boxes.append(BoundBox(x=x,y=y,z=z,h=0,w=0))
            f = Filament(boxes)
            f.meta["fid"] = id
            fils.append(f)
        return fils

    def filaments_to_coords(self, filaments : List[Filament]) -> np.array:

        Npos = np.sum([len(fil.boxes) for fil in filaments])

        coords_array = np.zeros(shape=(Npos,4))

        pos_counter = 0

        for fid, fil in enumerate(filaments):
            for box in fil.boxes:
                coords_array[pos_counter, 0] = box.x
                coords_array[pos_counter, 1] = box.y
                coords_array[pos_counter, 2] = box.z
                coords_array[pos_counter, 3] = fid + 1
                pos_counter = pos_counter + 1

        return coords_array



    def is_valid(self, data : np.array) -> bool:
        """
        Checks if it is valid .coords file
        :param data:
        :return:
        """
        return data.shape[1] == 4


    def get_data(self, files : List[str]) -> Dict[str, List[Filament]]:
        datasets = {}
        # Get data
        for file in files:
            data = np.genfromtxt(file)

            if self.is_valid(data) == False:
                continue

            filaments = self.coords_to_filaments(data)
            datasets[file] = filaments

        return datasets

    def write_data(self, dataset : Dict[str, List[Filament]], output_folder : str):

        out_fid = os.path.join(output_folder,"COORDS_FID")
        out_nofid = os.path.join(output_folder,"COORDS")
        os.makedirs(out_fid)
        os.makedirs(out_nofid)

        for file in dataset:
            filaments = dataset[file]
            coords = self.filaments_to_coords(filaments)

            filename = os.path.basename(file)
            outpath_fid = os.path.join(out_fid,filename)
            outpath_nofid = os.path.join(out_nofid, filename.replace("_fid",""))

            np.savetxt(outpath_fid, coords, fmt='%f')
            np.savetxt(outpath_nofid, coords[:,:3], fmt='%f')





    def do_resampling(self, data : Dict[str, List[Filament]], box_distance : float) -> Dict[str, List[Filament]]:

        resampeld_filaments = {}
        resampler = FilamentResampler(box_distance=box_distance)
        for file in data:
            filaments = data[file]
            filaments = resampler.resampling(filaments)
            resampeld_filaments[file] = filaments
        return resampeld_filaments

    def run(self, args):
        import os
        import glob
        input_dir = args.input
        if os.path.isfile(input_dir):
            files = [input_dir]
        else:
            path = os.path.join(os.path.abspath(input_dir), "*.coords")
            files = glob.glob(path)
        new_distance = args.distance
        outputpath = args.output

        print("Load data")
        datasets = self.get_data(files)
        print(datasets)
        print("Resample data")
        datasets_resampeld = self.do_resampling(data=datasets,box_distance=new_distance)

        print("Write data")
        self.write_data(datasets_resampeld,output_folder=outputpath)

        print("Done")






